"""The Winston-Lutz module loads and processes EPID images that have acquired Winston-Lutz type images.

Features:

* **Couch shift instructions** - After running a WL test, get immediate feedback on how to shift the couch.
  Couch values can also be passed in and the new couch values will be presented so you don't have to do that pesky conversion.
  "Do I subtract that number or add it?"
* **Automatic field & BB positioning** - When an image or directory is loaded, the field CAX and the BB
  are automatically found, along with the vector and scalar distance between them.
* **Isocenter size determination** - Using backprojections of the EPID images, the 3D gantry isocenter size
  and position can be determined *independent of the BB position*. Additionally, the 2D planar isocenter size
  of the collimator and couch can also be determined.
* **Image plotting** - WL images can be plotted separately or together, each of which shows the field CAX, BB and
  scalar distance from BB to CAX.
* **Axis deviation plots** - Plot the variation of the gantry, collimator, couch, and EPID in each plane
  as well as RMS variation.
* **File name interpretation** - Rename DICOM filenames to include axis information for linacs that don't include
  such information in the DICOM tags. E.g. "myWL_gantry45_coll0_couch315.dcm".
"""
import dataclasses
import enum
import io
import math
import os.path as osp
import webbrowser
from dataclasses import dataclass
from itertools import zip_longest
from pathlib import Path
from textwrap import wrap
from typing import Union, List, Tuple, Optional, BinaryIO, Iterable, Dict

import argue
import matplotlib.pyplot as plt
import numpy as np
from scipy import ndimage, optimize, linalg
from skimage import measure
from skimage.measure._regionprops import RegionProperties

from .core import image, pdf
from .core.decorators import lru_cache
from .core.geometry import Point, Line, Vector, cos, sin
from .core.io import TemporaryZipDirectory, get_url, retrieve_demo_file, is_dicom_image
from .core.mask import bounding_box
from .core.utilities import is_close, ResultBase, convert_to_enum


class Axis(enum.Enum):
    GANTRY = "Gantry"  #:
    COLLIMATOR = "Collimator"  #:
    COUCH = "Couch"  #:
    GB_COMBO = "GB Combo"  #:
    GBP_COMBO = "GBP Combo"  #:
    EPID = "Epid"  #:
    REFERENCE = "Reference"  #:


@dataclass
class WinstonLutz2DResult(ResultBase):
    variable_axis: str  #:
    cax2epid_vector: Vector  #:
    cax2epid_distance: float  #:
    cax2bb_distance: float  #:
    cax2bb_vector: Vector  #:
    bb_location: Point  #:
    field_cax: Point  #:


@dataclass
class WinstonLutzResult(ResultBase):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    num_gantry_images: int  #:
    num_gantry_coll_images: int  #:
    num_coll_images: int  #:
    num_couch_images: int  #:
    num_total_images: int  #:
    max_2d_cax_to_bb_mm: float  #:
    median_2d_cax_to_bb_mm: float  #:
    max_2d_cax_to_epid_mm: float  #:
    median_2d_cax_to_epid_mm: float  #:
    gantry_3d_iso_diameter_mm: float  #:
    max_gantry_rms_deviation_mm: float  #:
    max_epid_rms_deviation_mm: float  #:
    gantry_coll_3d_iso_diameter_mm: float  #:
    coll_2d_iso_diameter_mm: float  #:
    max_coll_rms_deviation_mm: float  #:
    couch_2d_iso_diameter_mm: float  #:
    max_couch_rms_deviation_mm: float  #:


class WinstonLutz:
    """Class for performing a Winston-Lutz test of the radiation isocenter."""

    images: list  #:

    def __init__(
        self,
        directory: Union[str, List[str], Path],
        use_filenames: bool = False,
        axis_mapping: Optional[Dict[str, Tuple[int, int, int]]] = None,
    ):
        """
        Parameters
        ----------
        directory : str, list[str]
            Path to the directory of the Winston-Lutz EPID images or a list of the image paths
        use_filenames: bool
            Whether to try to use the file name to determine axis values.
            Useful for Elekta machines that do not include that info in the DICOM data.
        axis_mapping: dict
            An optional way of instatiating by passing each file along with the axis values.
            Structure should be <filename>: (<gantry>, <coll>, <couch>).
        """
        self.images = []
        if axis_mapping:
            for filename, (gantry, coll, couch) in axis_mapping.items():
                self.images.append(
                    WinstonLutz2D(
                        Path(directory) / filename,
                        use_filenames=False,
                        gantry=gantry,
                        coll=coll,
                        couch=couch,
                    )
                )
        elif isinstance(directory, list):
            for file in directory:
                if is_dicom_image(file):
                    img = WinstonLutz2D(file, use_filenames)
                    self.images.append(img)
        elif not osp.isdir(directory):
            raise ValueError(
                "Invalid directory passed. Check the correct method and file was used."
            )
        else:
            image_files = image.retrieve_image_files(directory)
            for file in image_files:
                img = WinstonLutz2D(file, use_filenames)
                self.images.append(img)
        if len(self.images) < 2:
            raise ValueError(
                "<2 valid WL images were found in the folder/file or passed. Ensure you chose the correct folder/file for analysis."
            )
        self.images.sort(
            key=lambda i: (i.gantry_angle, i.collimator_angle, i.couch_angle)
        )
        self._is_analyzed = False

    @classmethod
    def from_demo_images(cls):
        """Instantiate using the demo images."""
        demo_file = retrieve_demo_file(name="winston_lutz.zip")
        return cls.from_zip(demo_file)

    @classmethod
    def from_zip(cls, zfile: Union[str, BinaryIO], use_filenames: bool = False):
        """Instantiate from a zip file rather than a directory.

        Parameters
        ----------
        zfile
            Path to the archive file.
        use_filenames : bool
            Whether to interpret axis angles using the filenames.
            Set to true for Elekta machines where the gantry/coll/couch data is not in the DICOM metadata.
        """
        with TemporaryZipDirectory(zfile) as tmpz:
            obj = cls(tmpz, use_filenames=use_filenames)
        return obj

    @classmethod
    def from_url(cls, url: str, use_filenames: bool = False):
        """Instantiate from a URL.

        Parameters
        ----------
        url : str
            URL that points to a zip archive of the DICOM images.
        use_filenames : bool
            Whether to interpret axis angles using the filenames.
            Set to true for Elekta machines where the gantry/coll/couch data is not in the DICOM metadata.
        """
        zfile = get_url(url)
        return cls.from_zip(zfile, use_filenames=use_filenames)

    @staticmethod
    def run_demo():
        """Run the Winston-Lutz demo, which loads the demo files, prints results, and plots a summary image."""
        wl = WinstonLutz.from_demo_images()
        wl.analyze()
        print(wl.results())
        wl.plot_summary()

    def analyze(self, bb_size_mm: float = 5):
        """Analyze the WL images.

        Parameters
        ----------
        bb_size_mm
            The expected size of the BB in mm. The actual size of the BB can be +/-2mm from the passed value.
        """
        for img in self.images:
            img.analyze(bb_size_mm)
        self._is_analyzed = True

    @lru_cache()
    def _minimize_axis(self, axes: Union[Axis, Tuple[Axis, ...]] = (Axis.GANTRY,)):
        """Return the minimization result of the given axis."""
        if isinstance(axes, Axis):
            axes = (axes,)

        def max_distance_to_lines(p, lines) -> float:
            """Calculate the maximum distance to any line from the given point."""
            point = Point(p[0], p[1], p[2])
            return max(line.distance_to(point) for line in lines)

        things = [
            image.cax_line_projection
            for image in self.images
            if image.variable_axis in (axes + (Axis.REFERENCE,))
        ]
        if len(things) <= 1:
            raise ValueError(
                "Not enough images of the given type to identify the axis isocenter"
            )
        initial_guess = np.array([0, 0, 0])
        bounds = [(-20, 20), (-20, 20), (-20, 20)]
        result = optimize.minimize(
            max_distance_to_lines, initial_guess, args=things, bounds=bounds
        )
        return result

    @property
    def gantry_iso_size(self) -> float:
        """The diameter of the 3D gantry isocenter size in mm. Only images where the collimator
        and couch were at 0 are used to determine this value."""
        num_gantry_like_images = self._get_images((Axis.GANTRY, Axis.REFERENCE))[0]
        if num_gantry_like_images > 1:
            return self._minimize_axis(Axis.GANTRY).fun * 2
        else:
            return 0

    @property
    def gantry_coll_iso_size(self) -> float:
        """The diameter of the 3D gantry isocenter size in mm *including collimator and gantry/coll combo images*.
        Images where the couch!=0 are excluded."""
        num_gantry_like_images = self._get_images(
            (Axis.GANTRY, Axis.COLLIMATOR, Axis.GB_COMBO, Axis.REFERENCE)
        )[0]
        if num_gantry_like_images > 1:
            return (
                self._minimize_axis((Axis.GANTRY, Axis.COLLIMATOR, Axis.GB_COMBO)).fun
                * 2
            )
        else:
            return 0

    @staticmethod
    def _find_max_distance_between_points(images) -> float:
        """Find the maximum distance between a set of points. Used for 2D images like collimator and couch."""
        points = [
            Point(image.cax2bb_vector.x, image.cax2bb_vector.y) for image in images
        ]
        dists = []
        for point1 in points:
            for point2 in points:
                p = point1.distance_to(point2)
                dists.append(p)
        return max(dists)

    @property
    def collimator_iso_size(self) -> float:
        """The 2D collimator isocenter size (diameter) in mm. The iso size is in the plane
        normal to the gantry."""
        num_collimator_like_images, images = self._get_images(
            (Axis.COLLIMATOR, Axis.REFERENCE)
        )
        if num_collimator_like_images > 1:
            return self._find_max_distance_between_points(images)
        else:
            return 0

    @property
    def couch_iso_size(self) -> float:
        """The diameter of the 2D couch isocenter size in mm. Only images where
        the gantry and collimator were at zero are used to determine this value."""
        num_couch_like_images, images = self._get_images((Axis.COUCH, Axis.REFERENCE))
        if num_couch_like_images > 1:
            return self._find_max_distance_between_points(images)
        else:
            return 0

    @property
    def bb_shift_vector(self) -> Vector:
        """The shift necessary to place the BB at the radiation isocenter.
        The values are in the coordinates defined in the documentation.

        The shift is based on the paper by Low et al. See online documentation for more.
        """
        A = np.empty([2 * len(self.images), 3])
        epsilon = np.empty([2 * len(self.images), 1])
        for idx, img in enumerate(self.images):
            g = img.gantry_angle
            c = img.couch_angle_varian_scale
            A[2 * idx : 2 * idx + 2, :] = np.array(
                [
                    [-cos(c), -sin(c), 0],
                    [-cos(g) * sin(c), cos(g) * cos(c), -sin(g)],
                ]
            )  # equation 6 (minus delta)
            epsilon[2 * idx : 2 * idx + 2] = np.array(
                [[img.cax2bb_vector.y], [img.cax2bb_vector.x]]
            )  # equation 7

        B = linalg.pinv(A)
        delta = B.dot(epsilon)  # equation 9
        return Vector(x=delta[1][0], y=-delta[0][0], z=-delta[2][0])

    def bb_shift_instructions(
        self,
        couch_vrt: Optional[float] = None,
        couch_lng: Optional[float] = None,
        couch_lat: Optional[float] = None,
    ) -> str:
        """Returns a string describing how to shift the BB to the radiation isocenter looking from the foot of the couch.
        Optionally, the current couch values can be passed in to get the new couch values. If passing the current
        couch position all values must be passed.

        Parameters
        ----------
        couch_vrt : float
            The current couch vertical position in cm.
        couch_lng : float
            The current couch longitudinal position in cm.
        couch_lat : float
            The current couch lateral position in cm.
        """
        sv = self.bb_shift_vector
        x_dir = "LEFT" if sv.x < 0 else "RIGHT"
        y_dir = "IN" if sv.y > 0 else "OUT"
        z_dir = "UP" if sv.z > 0 else "DOWN"
        move = f"{x_dir} {abs(sv.x):2.2f}mm; {y_dir} {abs(sv.y):2.2f}mm; {z_dir} {abs(sv.z):2.2f}mm"
        if all(val is not None for val in [couch_vrt, couch_lat, couch_lng]):
            new_lat = round(couch_lat + sv.x / 10, 2)
            new_vrt = round(couch_vrt + sv.z / 10, 2)
            new_lng = round(couch_lng + sv.y / 10, 2)
            move += f"\nNew couch coordinates (mm): VRT: {new_vrt:3.2f}; LNG: {new_lng:3.2f}; LAT: {new_lat:3.2f}"
        return move

    @argue.options(value=("all", "range"))
    def axis_rms_deviation(
        self, axis: Axis = Axis.GANTRY, value: str = "all"
    ) -> Union[Iterable, float]:
        """The RMS deviations of a given axis/axes.

        Parameters
        ----------
        axis : ('Gantry', 'Collimator', 'Couch', 'Epid', 'GB Combo',  'GBP Combo')
            The axis desired.
        value : {'all', 'range'}
            Whether to return all the RMS values from all images for that axis, or only return the maximum range of
            values, i.e. the 'sag'.
        """
        axis = convert_to_enum(axis, Axis)
        if axis != Axis.EPID:
            attr = "cax2bb_vector"
        else:
            attr = "cax2epid_vector"
            axis = (Axis.GANTRY, Axis.COLLIMATOR, Axis.REFERENCE)
        imgs = self._get_images(axis=axis)[1]
        if len(imgs) <= 1:
            return (0,)
        rms = [getattr(img, attr).as_scalar() for img in imgs]
        if value == "range":
            rms = max(rms) - min(rms)
        return rms

    @argue.options(metric=("max", "median"))
    def cax2bb_distance(self, metric: str = "max") -> float:
        """The distance in mm between the CAX and BB for all images according to the given metric.

        Parameters
        ----------
        metric : {'max', 'median'}
            The metric of distance to use.
        """
        if metric == "max":
            return max(image.cax2bb_distance for image in self.images)
        elif metric == "median":
            return np.median([image.cax2bb_distance for image in self.images])

    @argue.options(metric=("max", "median"))
    def cax2epid_distance(self, metric: str = "max") -> float:
        """The distance in mm between the CAX and EPID center pixel for all images according to the given metric.

        Parameters
        ----------
        metric : {'max', 'median'}
            The metric of distance to use.
        """
        if metric == "max":
            return max(image.cax2epid_distance for image in self.images)
        elif metric == "median":
            return np.median([image.cax2epid_distance for image in self.images])

    def _plot_deviation(
        self, axis: Axis, ax: Optional[plt.Axes] = None, show: bool = True
    ):
        """Helper function: Plot the sag in Cartesian coordinates.

        Parameters
        ----------
        axis : {'gantry', 'epid', 'collimator', 'couch'}
            The axis to plot.
        ax : None, matplotlib.Axes
            The axis to plot to. If None, creates a new plot.
        show : bool
            Whether to show the image.
        """
        title = f"In-plane {axis.value} displacement"
        if axis == Axis.EPID:
            attr = "cax2epid_vector"
            axis = Axis.GANTRY
        else:
            attr = "cax2bb_vector"
        # get axis images, angles, and shifts
        imgs = [
            image
            for image in self.images
            if image.variable_axis in (axis, Axis.REFERENCE)
        ]
        angles = [
            getattr(image, "{}_angle".format(axis.value.lower())) for image in imgs
        ]
        xz_sag = np.array([getattr(img, attr).x for img in imgs])
        y_sag = np.array([getattr(img, attr).y for img in imgs])
        rms = np.sqrt(xz_sag**2 + y_sag**2)

        # plot the axis deviation
        if ax is None:
            ax = plt.subplot(111)
        ax.plot(angles, y_sag, "bo", label="Y-axis", ls="-.")
        ax.plot(angles, xz_sag, "m^", label="X/Z-axis", ls="-.")
        ax.plot(angles, rms, "g+", label="RMS", ls="-")
        ax.set_title(title)
        ax.set_ylabel("mm")
        ax.set_xlabel(f"{axis.value} angle")
        ax.set_xticks(np.arange(0, 361, 45))
        ax.set_xlim(-15, 375)
        ax.grid(True)
        ax.legend(numpoints=1)
        if show:
            plt.show()

    def _get_images(
        self, axis: Union[Axis, Tuple[Axis, ...]] = (Axis.GANTRY,)
    ) -> Tuple[float, list]:
        if isinstance(axis, Axis):
            axis = (axis,)
        images = [image for image in self.images if image.variable_axis in axis]
        return len(images), images

    def plot_axis_images(
        self, axis: Axis = Axis.GANTRY, show: bool = True, ax: Optional[plt.Axes] = None
    ):
        """Plot all CAX/BB/EPID positions for the images of a given axis.

        For example, axis='Couch' plots a reference image, and all the BB points of the other
        images where the couch was moving.

        Parameters
        ----------
        axis : {'Gantry', 'Collimator', 'Couch', 'GB Combo',  'GBP Combo'}
            The images/markers from which accelerator axis to plot.
        show : bool
            Whether to actually show the images.
        ax : None, matplotlib.Axes
            The axis to plot to. If None, creates a new plot.
        """
        axis = convert_to_enum(axis, Axis)
        images = [
            image
            for image in self.images
            if image.variable_axis in (axis, Axis.REFERENCE)
        ]
        ax = images[0].plot(
            show=False, ax=ax
        )  # plots the first marker; plot the rest of the markers below
        if axis != Axis.COUCH:
            # plot EPID
            epid_xs = [img.epid.x for img in images[1:]]
            epid_ys = [img.epid.y for img in images[1:]]
            ax.plot(epid_xs, epid_ys, "b+", ms=8)
            # get CAX positions
            xs = [img.field_cax.x for img in images[1:]]
            ys = [img.field_cax.y for img in images[1:]]
            marker = "gs"
        else:
            # get BB positions
            xs = [img.bb.x for img in images[1:]]
            ys = [img.bb.y for img in images[1:]]
            marker = "ro"
        ax.plot(xs, ys, marker, ms=8)
        # set labels
        ax.set_title(axis.value + " wobble")
        ax.set_xlabel(axis.value + " positions superimposed")
        ax.set_ylabel(
            axis.value
            + f" iso size: {getattr(self, axis.value.lower() + '_iso_size'):3.2f}mm"
        )
        if show:
            plt.show()

    def plot_images(
        self, axis: Axis = Axis.GANTRY, show: bool = True, split: bool = False, **kwargs
    ) -> (List[plt.Figure], List[str]):
        """Plot a grid of all the images acquired.

        Four columns are plotted with the titles showing which axis that column represents.

        Parameters
        ----------
        axis : {'Gantry', 'Collimator', 'Couch', 'GB Combo', 'GBP Combo', 'All'}
        show : bool
            Whether to show the image.
        split : bool
            Whether to show/plot the images individually or as one large figure.
        """
        axis = convert_to_enum(axis, Axis)
        if not self._is_analyzed:
            raise ValueError("The set is not analyzed. Use .analyze() first.")

        def plot_image(image, axis):
            """Helper function to plot a WLImage to an axis."""
            if image is None:
                axis.set_frame_on(False)
                axis.axis("off")
            else:
                image.plot(ax=axis, show=False)

        # get axis images
        if axis == Axis.GANTRY:
            images = [
                image
                for image in self.images
                if image.variable_axis in (Axis.GANTRY, Axis.REFERENCE)
            ]
        elif axis == Axis.COLLIMATOR:
            images = [
                image
                for image in self.images
                if image.variable_axis in (Axis.COLLIMATOR, Axis.REFERENCE)
            ]
        elif axis == Axis.COUCH:
            images = [
                image
                for image in self.images
                if image.variable_axis in (Axis.COUCH, Axis.REFERENCE)
            ]
        elif axis == Axis.GB_COMBO:
            images = [
                image
                for image in self.images
                if image.variable_axis
                in (Axis.GB_COMBO, Axis.GANTRY, Axis.COLLIMATOR, Axis.REFERENCE)
            ]
        elif axis == Axis.GBP_COMBO:
            images = self.images

        # set the figsize if it wasn't passed
        if not kwargs.get("figsize"):
            dpi = 72
            width_px = 1080
            width_in = width_px / dpi
            if not split:
                max_num_images = math.ceil(len(images) / 4)
                height_in = (width_in / 4) * max_num_images
            else:
                height_in = width_in = 3
            kwargs["figsize"] = (width_in, height_in)

        figs = []
        names = []
        # create plots
        if not split:
            fig, axes = plt.subplots(nrows=max_num_images, ncols=4, **kwargs)
            for mpl_axis, wl_image in zip_longest(axes.flatten(), images):
                plot_image(wl_image, mpl_axis)

            # set titles
            fig.suptitle(f"{axis.value} images", fontsize=14, y=1)
            fig.tight_layout()
            figs.append(fig)
            names.append("image")
        else:
            for wl_image in images:
                fig, axes = plt.subplots(**kwargs)
                plot_image(wl_image, axes)
                figs.append(fig)
                names.append(str(wl_image))

        if show:
            plt.show()

        return figs, names

    def save_images(
        self, filename: Union[str, BinaryIO], axis: Axis = Axis.GANTRY, **kwargs
    ):
        """Save the figure of `plot_images()` to file. Keyword arguments are passed to `matplotlib.pyplot.savefig()`.

        Parameters
        ----------
        filename : str
            The name of the file to save to.
        axis
            The axis to save.
        """
        self.plot_images(axis=axis, show=False)
        plt.savefig(filename, **kwargs)

    def save_images_to_stream(self, **kwargs) -> Dict[str, io.BytesIO]:
        """Save the individual image plots to stream"""
        figs, names = self.plot_images(
            axis=Axis.GBP_COMBO, show=False, split=True
        )  # all images
        streams = [io.BytesIO() for _ in figs]
        for fig, stream in zip(figs, streams):
            fig.savefig(stream, **kwargs)
        return {name: stream for name, stream in zip(names, streams)}

    def plot_summary(self, show: bool = True, fig_size: Optional[tuple] = None):
        """Plot a summary figure showing the gantry sag and wobble plots of the three axes."""
        if not self._is_analyzed:
            raise ValueError("The set is not analyzed. Use .analyze() first.")
        figsize = (11, 9) if fig_size is None else fig_size
        plt.figure(figsize=figsize)
        grid = (3, 6)
        gantry_sag_ax = plt.subplot2grid(grid, (0, 0), colspan=3)
        self._plot_deviation(Axis.GANTRY, gantry_sag_ax, show=False)
        epid_sag_ax = plt.subplot2grid(grid, (0, 3), colspan=3)
        self._plot_deviation(Axis.EPID, epid_sag_ax, show=False)
        if self._get_images((Axis.COLLIMATOR, Axis.REFERENCE))[0] > 1:
            coll_sag_ax = plt.subplot2grid(grid, (1, 0), colspan=3)
            self._plot_deviation(Axis.COLLIMATOR, coll_sag_ax, show=False)
        if self._get_images((Axis.COUCH, Axis.REFERENCE))[0] > 1:
            couch_sag_ax = plt.subplot2grid(grid, (1, 3), colspan=3)
            self._plot_deviation(Axis.COUCH, couch_sag_ax, show=False)

        for axis, axnum in zip((Axis.GANTRY, Axis.COLLIMATOR, Axis.COUCH), (0, 2, 4)):
            if self._get_images((axis, Axis.REFERENCE))[0] > 1:
                ax = plt.subplot2grid(grid, (2, axnum), colspan=2)
                self.plot_axis_images(axis=axis, ax=ax, show=False)
        if show:
            plt.tight_layout()
            plt.show()

    def save_summary(self, filename: Union[str, BinaryIO], **kwargs):
        """Save the summary image."""
        self.plot_summary(show=False, fig_size=kwargs.pop("fig_size", None))
        plt.tight_layout()
        plt.savefig(filename, **kwargs)

    def results(self, as_list: bool = False) -> str:
        """Return the analysis results summary.

        Parameters
        ----------
        as_list : bool
            Whether to return as a list of strings vs single string. Pretty much for internal usage.
        """
        if not self._is_analyzed:
            raise ValueError("The set is not analyzed. Use .analyze() first.")
        num_gantry_imgs = self._get_images(axis=(Axis.GANTRY, Axis.REFERENCE))[0]
        num_gantry_coll_imgs = self._get_images(
            axis=(Axis.GANTRY, Axis.COLLIMATOR, Axis.GB_COMBO, Axis.REFERENCE)
        )[0]
        num_coll_imgs = self._get_images(axis=(Axis.COLLIMATOR, Axis.REFERENCE))[0]
        num_couch_imgs = self._get_images(axis=(Axis.COUCH, Axis.REFERENCE))[0]
        num_imgs = len(self.images)
        result = [
            "Winston-Lutz Analysis",
            "=================================",
            f"Number of images: {num_imgs}",
            f"Maximum 2D CAX->BB distance: {self.cax2bb_distance('max'):.2f}mm",
            f"Median 2D CAX->BB distance: {self.cax2bb_distance('median'):.2f}mm",
            f"Shift to iso: facing gantry, move BB: {self.bb_shift_instructions()}",
            f"Gantry 3D isocenter diameter: {self.gantry_iso_size:.2f}mm ({num_gantry_imgs}/{num_imgs} images considered)",
            f"Maximum Gantry RMS deviation (mm): {max(self.axis_rms_deviation(Axis.GANTRY)):.2f}mm",
            f"Maximum EPID RMS deviation (mm): {max(self.axis_rms_deviation(Axis.EPID)):.2f}mm",
            f"Gantry+Collimator 3D isocenter diameter: {self.gantry_coll_iso_size:.2f}mm ({num_gantry_coll_imgs}/{num_imgs} images considered)",
            f"Collimator 2D isocenter diameter: {self.collimator_iso_size:.2f}mm ({num_coll_imgs}/{num_imgs} images considered)",
            f"Maximum Collimator RMS deviation (mm): {max(self.axis_rms_deviation(Axis.COLLIMATOR)):.2f}",
            f"Couch 2D isocenter diameter: {self.couch_iso_size:.2f}mm ({num_couch_imgs}/{num_imgs} images considered)",
            f"Maximum Couch RMS deviation (mm): {max(self.axis_rms_deviation(Axis.COUCH)):.2f}",
        ]
        if not as_list:
            result = "\n".join(result)
        return result

    def results_data(self, as_dict=False) -> Union[WinstonLutzResult, dict]:
        """Present the results data and metadata as a dataclass or dict.
        The default return type is a dataclass."""
        if not self._is_analyzed:
            raise ValueError("The set is not analyzed. Use .analyze() first.")
        num_gantry_imgs = self._get_images(axis=(Axis.GANTRY, Axis.REFERENCE))[0]
        num_gantry_coll_imgs = self._get_images(
            axis=(Axis.GANTRY, Axis.COLLIMATOR, Axis.GB_COMBO, Axis.REFERENCE)
        )[0]
        num_coll_imgs = self._get_images(axis=(Axis.COLLIMATOR, Axis.REFERENCE))[0]
        num_couch_imgs = self._get_images(axis=(Axis.COUCH, Axis.REFERENCE))[0]

        data = WinstonLutzResult(
            num_total_images=len(self.images),
            num_gantry_images=num_gantry_imgs,
            num_coll_images=num_coll_imgs,
            num_gantry_coll_images=num_gantry_coll_imgs,
            num_couch_images=num_couch_imgs,
            max_2d_cax_to_bb_mm=self.cax2bb_distance("max"),
            median_2d_cax_to_bb_mm=self.cax2bb_distance("median"),
            max_2d_cax_to_epid_mm=self.cax2epid_distance("max"),
            median_2d_cax_to_epid_mm=self.cax2epid_distance("median"),
            coll_2d_iso_diameter_mm=self.collimator_iso_size,
            couch_2d_iso_diameter_mm=self.couch_iso_size,
            gantry_3d_iso_diameter_mm=self.gantry_iso_size,
            gantry_coll_3d_iso_diameter_mm=self.gantry_coll_iso_size,
            max_gantry_rms_deviation_mm=max(self.axis_rms_deviation(axis=Axis.GANTRY)),
            max_coll_rms_deviation_mm=max(
                self.axis_rms_deviation(axis=Axis.COLLIMATOR)
            ),
            max_couch_rms_deviation_mm=max(self.axis_rms_deviation(axis=Axis.COUCH)),
            max_epid_rms_deviation_mm=max(self.axis_rms_deviation(axis=Axis.EPID)),
        )
        if as_dict:
            return dataclasses.asdict(data)
        return data

    def publish_pdf(
        self,
        filename: str,
        notes: Optional[Union[str, List[str]]] = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ):
        """Publish (print) a PDF containing the analysis, images, and quantitative results.

        Parameters
        ----------
        filename : (str, file-like object}
            The file to write the results to.
        notes : str, list of strings
            Text; if str, prints single line.
            If list of strings, each list item is printed on its own line.
        open_file : bool
            Whether to open the file using the default program after creation.
        metadata : dict
            Extra data to be passed and shown in the PDF. The key and value will be shown with a colon.
            E.g. passing {'Author': 'James', 'Unit': 'TrueBeam'} would result in text in the PDF like:
            --------------
            Author: James
            Unit: TrueBeam
            --------------
        """
        if not self._is_analyzed:
            raise ValueError("The set is not analyzed. Use .analyze() first.")
        plt.ioff()
        title = "Winston-Lutz Analysis"
        canvas = pdf.PylinacCanvas(filename, page_title=title, metadata=metadata)
        text = self.results(as_list=True)
        canvas.add_text(text=text, location=(7, 25.5))
        # draw summary image on 1st page
        data = io.BytesIO()
        self.save_summary(data, fig_size=(8, 8))
        canvas.add_image(image_data=data, location=(2, 3), dimensions=(16, 16))
        if notes is not None:
            canvas.add_text(text="Notes:", location=(1, 4.5), font_size=14)
            canvas.add_text(text=notes, location=(1, 4))
        # add more pages showing individual axis images
        for ax in (
            Axis.GANTRY,
            Axis.COLLIMATOR,
            Axis.COUCH,
            Axis.GB_COMBO,
            Axis.GBP_COMBO,
        ):
            if self._contains_axis_images(ax):
                canvas.add_new_page()
                data = io.BytesIO()
                self.save_images(data, axis=ax)
                canvas.add_image(data, location=(2, 7), dimensions=(18, 18))

        canvas.finish()

        if open_file:
            webbrowser.open(filename)

    def _contains_axis_images(self, axis: Axis = Axis.GANTRY) -> bool:
        """Return whether or not the set of WL images contains images pertaining to a given axis"""
        return any(True for image in self.images if image.variable_axis in (axis,))


class WinstonLutz2D(image.LinacDicomImage):
    """Holds individual Winston-Lutz EPID images, image properties, and automatically finds the field CAX and BB."""

    bb: Point
    field_cax: Point
    _rad_field_bounding_box: list

    def __init__(
        self, file: Union[str, BinaryIO, Path], use_filenames: bool = False, **kwargs
    ):
        """
        Parameters
        ----------
        file : str
            Path to the image file.
        use_filenames: bool
            Whether to try to use the file name to determine axis values.
            Useful for Elekta machines that do not include that info in the DICOM data.
        """
        super().__init__(file, use_filenames=use_filenames, **kwargs)
        self._is_analyzed = False

    def analyze(self, bb_size_mm: float = 5) -> None:
        """Analyze the image."""
        self.check_inversion_by_histogram(percentiles=(0.01, 50, 99.99))
        self.flipud()
        self._clean_edges()
        self.ground()
        self.normalize()
        self.field_cax, self._rad_field_bounding_box = self._find_field_centroid()
        self.bb = self._find_bb(bb_size_mm)
        self._is_analyzed = True

    def __repr__(self):
        return f"WLImage(G={self.gantry_angle:.1f}, B={self.collimator_angle:.1f}, P={self.couch_angle:.1f})"

    def _clean_edges(self, window_size: int = 2) -> None:
        """Clean the edges of the image to be near the background level."""

        def has_noise(self, window_size):
            """Helper method to determine if there is spurious signal at any of the image edges.

            Determines if the min or max of an edge is within 10% of the baseline value and trims if not.
            """
            near_min, near_max = np.percentile(self.array, [5, 99.5])
            img_range = near_max - near_min
            top = self[:window_size, :]
            left = self[:, :window_size]
            bottom = self[-window_size:, :]
            right = self[:, -window_size:]
            edge_array = np.concatenate(
                (top.flatten(), left.flatten(), bottom.flatten(), right.flatten())
            )
            edge_too_low = edge_array.min() < (near_min - img_range / 10)
            edge_too_high = edge_array.max() > (near_max + img_range / 10)
            return edge_too_low or edge_too_high

        safety_stop = np.min(self.shape) / 10
        while has_noise(self, window_size) and safety_stop > 0:
            self.crop(window_size)
            safety_stop -= 1

    def _find_field_centroid(self) -> Tuple[Point, List]:
        """Find the centroid of the radiation field based on a 50% height threshold.

        Returns
        -------
        p
            The CAX point location.
        edges
            The bounding box of the field, plus a small margin.
        """
        min, max = np.percentile(self.array, [5, 99.9])
        threshold_img = self.as_binary((max - min) / 2 + min)
        filled_img = ndimage.binary_fill_holes(threshold_img)
        # clean single-pixel noise from outside field
        cleaned_img = ndimage.binary_erosion(threshold_img)
        [*edges] = bounding_box(cleaned_img)
        edges[0] -= 10
        edges[1] += 10
        edges[2] -= 10
        edges[3] += 10
        coords = ndimage.measurements.center_of_mass(filled_img)
        p = Point(x=coords[-1], y=coords[0])
        return p, edges

    def _find_bb(self, bb_size: float) -> Point:
        """Find the BB within the radiation field. Iteratively searches for a circle-like object
        by lowering a low-pass threshold value until found.

        Returns
        -------
        Point
            The weighted-pixel value location of the BB.
        """
        # get initial starting conditions
        hmin, hmax = np.percentile(self.array, [5, 99.99])
        spread = hmax - hmin
        max_thresh = hmax
        lower_thresh = hmax - spread / 1.5
        # search for the BB by iteratively lowering the low-pass threshold value until the BB is found.
        found = False
        while not found:
            try:
                binary_arr = np.logical_and((max_thresh > self), (self >= lower_thresh))
                # use below for debugging
                # plt.imshow(binary_arr)
                # plt.show()
                labeled_arr, num_roi = ndimage.measurements.label(binary_arr)
                roi_sizes, bin_edges = np.histogram(labeled_arr, bins=num_roi + 1)
                bw_bb_img = np.where(
                    labeled_arr == np.argsort(roi_sizes)[-3], 1, 0
                )  # we pick the 3rd largest one because the largest is the background, 2nd is rad field, 3rd is the BB
                bw_bb_img = ndimage.binary_fill_holes(bw_bb_img).astype(
                    int
                )  # fill holes for low energy beams like 2.5MV
                bb_regionprops = measure.regionprops(bw_bb_img)[0]

                if not is_round(bb_regionprops):
                    raise ValueError
                if not is_modest_size(bw_bb_img, self.dpmm, bb_size):
                    raise ValueError
                if not is_symmetric(bw_bb_img):
                    raise ValueError
                if not is_near_center(bb_regionprops, self.dpmm, bw_bb_img.shape):
                    raise ValueError
            except (IndexError, ValueError):
                max_thresh -= 0.03 * spread
                if max_thresh < hmin:
                    raise ValueError(
                        "Unable to locate the BB. Make sure the field edges do not obscure the BB and that there is no artifacts in the images."
                    )
            else:
                found = True

        # determine the center of mass of the BB
        inv_img = image.load(self.array)
        # we invert so BB intensity increases w/ attenuation
        inv_img.check_inversion_by_histogram(percentiles=(99.99, 50, 0.01))
        bb_rprops = measure.regionprops(bw_bb_img, intensity_image=inv_img)[0]
        return Point(bb_rprops.weighted_centroid[1], bb_rprops.weighted_centroid[0])

    @property
    def epid(self) -> Point:
        """Center of the EPID panel"""
        return self.center

    @property
    def cax_line_projection(self) -> Line:
        """The projection of the field CAX through space around the area of the BB.
        Used for determining gantry isocenter size.

        Returns
        -------
        Line
            The virtual line in space made by the beam CAX.
        """
        p1 = Point()
        p2 = Point()
        # point 1 - ray origin
        p1.x = self.cax2bb_vector.x * cos(self.gantry_angle) + 20 * sin(
            self.gantry_angle
        )
        p1.y = self.cax2bb_vector.x * -sin(self.gantry_angle) + 20 * cos(
            self.gantry_angle
        )
        p1.z = self.cax2bb_vector.y
        # point 2 - ray destination
        p2.x = self.cax2bb_vector.x * cos(self.gantry_angle) - 20 * sin(
            self.gantry_angle
        )
        p2.y = self.cax2bb_vector.x * -sin(self.gantry_angle) - 20 * cos(
            self.gantry_angle
        )
        p2.z = self.cax2bb_vector.y
        l = Line(p1, p2)
        return l

    @property
    def couch_angle_varian_scale(self) -> float:
        """The couch angle converted from IEC 61217 scale to "Varian" scale. Note that any new Varian machine uses 61217."""
        #  convert to Varian scale per Low paper scale
        if super().couch_angle > 250:
            return 2 * 270 - super().couch_angle
        else:
            return 180 - super().couch_angle

    @property
    def cax2bb_vector(self) -> Vector:
        """The vector in mm from the CAX to the BB."""
        dist = (self.bb - self.field_cax) / self.dpmm
        return Vector(dist.x, dist.y, dist.z)

    @property
    def cax2bb_distance(self) -> float:
        """The scalar distance in mm from the CAX to the BB."""
        dist = self.field_cax.distance_to(self.bb)
        return dist / self.dpmm

    @property
    def cax2epid_vector(self) -> Vector:
        """The vector in mm from the CAX to the EPID center pixel"""
        dist = (self.epid - self.field_cax) / self.dpmm
        return Vector(dist.x, dist.y, dist.z)

    @property
    def cax2epid_distance(self) -> float:
        """The scalar distance in mm from the CAX to the EPID center pixel"""
        return self.field_cax.distance_to(self.epid) / self.dpmm

    def plot(
        self, ax: Optional[plt.Axes] = None, show: bool = True, clear_fig: bool = False
    ):
        """Plot the image, zoomed-in on the radiation field, along with the detected
        BB location and field CAX location.

        Parameters
        ----------
        ax : None, matplotlib Axes instance
            The axis to plot to. If None, will create a new figure.
        show : bool
            Whether to actually show the image.
        clear_fig : bool
            Whether to clear the figure first before drawing.
        """
        ax = super().plot(ax=ax, show=False, clear_fig=clear_fig)
        ax.plot(self.field_cax.x, self.field_cax.y, "gs", ms=8)
        ax.plot(self.bb.x, self.bb.y, "ro", ms=8)
        ax.axvline(x=self.epid.x, color="b")
        ax.axhline(y=self.epid.y, color="b")
        ax.set_ylim([self._rad_field_bounding_box[0], self._rad_field_bounding_box[1]])
        ax.set_xlim([self._rad_field_bounding_box[2], self._rad_field_bounding_box[3]])
        ax.set_yticklabels([])
        ax.set_xticklabels([])
        ax.set_title("\n".join(wrap(str(self.path), 30)), fontsize=10)
        ax.set_xlabel(
            f"G={self.gantry_angle:.0f}, B={self.collimator_angle:.0f}, P={self.couch_angle:.0f}"
        )
        ax.set_ylabel(f"CAX to BB: {self.cax2bb_distance:3.2f}mm")
        if show:
            plt.show()
        return ax

    def save_plot(self, filename: str, **kwargs):
        """Save the image plot to file."""
        self.plot(show=False)
        plt.savefig(filename, **kwargs)

    @property
    def variable_axis(self) -> Axis:
        """The axis that is varying.

        There are five types of images:

        * Reference : All axes are at 0.
        * Gantry: All axes but gantry at 0.
        * Collimator : All axes but collimator at 0.
        * Couch : All axes but couch at 0.
        * Combo : More than one axis is not at 0.
        """
        G0 = is_close(self.gantry_angle, [0, 360])
        B0 = is_close(self.collimator_angle, [0, 360])
        P0 = is_close(self.couch_angle, [0, 360])
        if G0 and B0 and not P0:
            return Axis.COUCH
        elif G0 and P0 and not B0:
            return Axis.COLLIMATOR
        elif P0 and B0 and not G0:
            return Axis.GANTRY
        elif P0 and B0 and G0:
            return Axis.REFERENCE
        elif P0:
            return Axis.GB_COMBO
        else:
            return Axis.GBP_COMBO

    def results_data(self, as_dict=False) -> Union[WinstonLutz2DResult, dict]:
        """Present the results data and metadata as a dataclass or dict.
        The default return type is a dataclass."""
        if not self._is_analyzed:
            raise ValueError("The image is not analyzed. Use .analyze() first.")

        data = WinstonLutz2DResult(
            variable_axis=self.variable_axis.value,
            cax2bb_vector=self.cax2bb_vector,
            cax2epid_vector=self.cax2epid_vector,
            cax2bb_distance=self.cax2bb_distance,
            cax2epid_distance=self.cax2epid_distance,
            bb_location=self.bb,
            field_cax=self.field_cax,
        )
        if as_dict:
            return dataclasses.asdict(data)
        return data


def is_symmetric(logical_array: np.ndarray) -> bool:
    """Whether the binary object's dimensions are symmetric, i.e. a perfect circle. Used to find the BB."""
    ymin, ymax, xmin, xmax = bounding_box(logical_array)
    y = abs(ymax - ymin)
    x = abs(xmax - xmin)
    if x > max(y * 1.05, y + 3) or x < min(y * 0.95, y - 3):
        return False
    return True


def is_near_center(
    region: RegionProperties, dpmm: float, shape: Tuple[float, float]
) -> bool:
    """Whether the bb is <2cm from the center of the field"""
    extent_limit_mm = 20
    bottom, left, top, right = region.bbox
    bb_center_x = left + (right - left) / 2
    bb_center_y = bottom + (top - bottom) / 2
    x_lo_limit = shape[1] / 2 - dpmm * extent_limit_mm
    x_hi_limit = shape[1] / 2 + dpmm * extent_limit_mm
    is_bb_x_centered = x_lo_limit < bb_center_x < x_hi_limit
    y_lo_limit = shape[0] / 2 - dpmm * extent_limit_mm
    y_hi_limit = shape[0] / 2 + dpmm * extent_limit_mm
    is_bb_y_centered = y_lo_limit < bb_center_y < y_hi_limit
    return is_bb_x_centered and is_bb_y_centered


def is_modest_size(logical_array: np.ndarray, dpmm: float, bb_size: float) -> bool:
    """Decide whether the ROI is roughly the size of a BB; not noise and not an artifact. Used to find the BB."""
    bb_area = np.sum(logical_array) / (dpmm**2)
    bb_size = max((bb_size, 2.1))
    expected_bb_area = np.pi * (bb_size / 2) ** 2
    larger_bb_area = np.pi * ((bb_size + 2) / 2) ** 2
    smaller_bb_area = np.pi * ((bb_size - 2) / 2) ** 2
    # return bool(np.isclose(bb_area, expected_bb_area, rtol=0.6))
    return smaller_bb_area < bb_area < larger_bb_area


def is_round(region: RegionProperties) -> bool:
    """Decide if the ROI is circular in nature by testing the filled area vs bounding box. Used to find the BB."""
    expected_fill_ratio = np.pi / 4  # area of a circle inside a square
    actual_fill_ratio = region.filled_area / region.bbox_area
    return expected_fill_ratio * 1.2 > actual_fill_ratio > expected_fill_ratio * 0.8



import dataclasses
import io
import textwrap
import webbrowser
from io import BytesIO
from pathlib import Path
from typing import Union, Dict, Optional, List, Tuple

import numpy as np
import scipy.ndimage
from matplotlib import pyplot as plt

from .core import pdf
from .core.geometry import Line, Point
from .core.profile import SingleProfile, Interpolation
from .core.utilities import ResultBase
from .ct import (
    CTP404CP504,
    AIR,
    ThicknessROI,
    CTP486,
    CatPhanModule,
    CatPhanBase,
    rois_to_results,
)

UNIFORMITY_OFFSET_MM = -45
GEOMETRY_OFFSET_MM = 45
ACRYLIC = 120
POLY = -35
TEFLON = 990


@dataclasses.dataclass
class QuartHUModuleOutput:
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    offset: int
    roi_settings: dict
    rois: dict
    measured_slice_thickness_mm: float
    signal_to_noise: float
    contrast_to_noise: float


@dataclasses.dataclass
class QuartGeometryModuleOutput:
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    offset: int
    roi_settings: dict
    rois: dict
    distances: dict


@dataclasses.dataclass
class QuartUniformityModuleOutput:
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    offset: int
    roi_settings: dict
    rois: dict
    passed: bool


@dataclasses.dataclass
class QuartDVTResult(ResultBase):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    phantom_model: str  #:
    phantom_roll_deg: float  #:
    origin_slice: int  #:
    num_images: int  #:
    hu_module: QuartHUModuleOutput  #:
    uniformity_module: QuartUniformityModuleOutput  #:
    geometric_module: QuartGeometryModuleOutput  #:


class QuartHUModule(CTP404CP504):
    roi_dist_mm = 52.5
    roi_radius_mm = 6
    roi_settings = {
        "Air": {
            "value": AIR,
            "angle": -90,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
        },
        "Poly": {
            "value": POLY,
            "angle": 0,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
        },
        "Acrylic": {
            "value": ACRYLIC,
            "angle": 45,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
        },
        "Teflon": {
            "value": TEFLON,
            "angle": 180,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
        },
    }
    background_roi_settings = {}
    thickness_roi_height = 25
    thickness_roi_width = 15
    thickness_roi_distance_mm = 32
    thickness_roi_settings = {
        "Bottom": {
            "angle": 90,
            "width": thickness_roi_height,
            "height": thickness_roi_width,
            "distance": thickness_roi_distance_mm,
        },
        "Top": {
            "angle": -90,
            "width": thickness_roi_height,
            "height": thickness_roi_width,
            "distance": thickness_roi_distance_mm,
        },
    }

    def _setup_geometry_rois(self) -> None:
        # no geometry ROIs
        pass

    def _setup_thickness_rois(self) -> None:
        """We invert the thickness ROIs because they are air gaps, not high-density wires"""
        self.thickness_image.invert()
        for name, setting in self.thickness_roi_settings.items():
            self.thickness_rois[name] = ThicknessROI(
                self.thickness_image,
                setting["width_pixels"],
                setting["height_pixels"],
                setting["angle_corrected"],
                setting["distance_pixels"],
                self.phan_center,
            )

    @property
    def meas_slice_thickness(self) -> float:
        """The average slice thickness for the 4 wire measurements in mm."""
        INCLINATION_CORRECTION = 0.577  # per manual; tan(30)
        return np.mean(
            sorted(
                roi.wire_fwhm * self.mm_per_pixel * INCLINATION_CORRECTION
                for roi in self.thickness_rois.values()
            )
        ) / (1 + 2 * self.pad)

    @property
    def signal_to_noise(self) -> float:
        """Calculate the SNR based on the suggested procedure in the manual:
        SNR = (HU + 1000) / sigma,
        where HU is the mean HU of a chosen insert and sigma is the stdev of the HU insert.
        We choose to use the Polystyrene as the target HU insert"""
        return (self.rois["Poly"].pixel_value + 1000) / self.rois["Poly"].std

    @property
    def contrast_to_noise(self) -> float:
        """Calculate the CNR based on the suggested procedure in the manual:
        CNR = abs(HU_target - HU_background) / sigma,
        where HU_target is the mean HU of a chosen insert, HU_background is the mean HU of the background insert
        and sigma is the stdev of the HU background.
        We choose to use the Polystyrene as the target HU insert and Acrylic (base phantom material) as the background"""
        return (
            abs(self.rois["Poly"].pixel_value - self.rois["Acrylic"].pixel_value)
            / self.rois["Acrylic"].std
        )


class QuartUniformityModule(CTP486):
    """Class for analysis of the Uniformity slice of the CTP module. Measures 5 ROIs around the slice that
    should all be close to the same value.
    """

    common_name = "HU Uniformity"
    roi_dist_mm = 53
    roi_radius_mm = 10
    nominal_value = 120
    roi_settings = {
        "Top": {
            "value": nominal_value,
            "angle": -90,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
        },
        "Right": {
            "value": nominal_value,
            "angle": 0,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
        },
        "Bottom": {
            "value": nominal_value,
            "angle": 90,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
        },
        "Left": {
            "value": nominal_value,
            "angle": 180,
            "distance": roi_dist_mm,
            "radius": roi_radius_mm,
        },
        "Center": {
            "value": nominal_value,
            "angle": 0,
            "distance": 0,
            "radius": roi_radius_mm,
        },
    }


class QuartGeometryModule(CatPhanModule):
    """Class for analysis of the Uniformity slice of the CTP module. Measures 5 ROIs around the slice that
    should all be close to the same value.
    """

    attr_name = "geometry_module"
    common_name = "Geometric Distortion"
    profiles: dict

    def _setup_rois(self) -> None:
        self.profiles = {}
        img = (
            self.image.array.copy()
        )  # we copy so we don't overwrite the existing image pixels
        img = scipy.ndimage.median_filter(img, size=3)
        img = img - img.min()  # ground the profile
        # calculate horizontal
        data = img[int(self.phan_center.y), :]
        prof = SingleProfile(
            data, interpolation=Interpolation.NONE, dpmm=1 / self.mm_per_pixel
        )
        fwhm = prof.fwxm_data()
        line = Line(
            Point(fwhm["left index (rounded)"], self.phan_center.y),
            Point(fwhm["right index (rounded)"], self.phan_center.y),
        )
        self.profiles["horizontal"] = {
            "width (mm)": fwhm["width (exact) mm"],
            "line": line,
        }
        # calculate vertical
        data = img[:, int(self.phan_center.x)]
        prof = SingleProfile(
            data, interpolation=Interpolation.NONE, dpmm=1 / self.mm_per_pixel
        )
        fwhm = prof.fwxm_data()
        line = Line(
            Point(self.phan_center.x, fwhm["left index (rounded)"]),
            Point(self.phan_center.x, fwhm["right index (rounded)"]),
        )
        self.profiles["vertical"] = {
            "width (mm)": fwhm["width (exact) mm"],
            "line": line,
        }

    def plot_rois(self, axis: plt.Axes):
        for name, profile_data in self.profiles.items():
            profile_data["line"].plot2axes(axis, width=2, color="blue")

    def distances(self) -> Dict[str, float]:
        """The measurements of the phantom size for the two lines in mm"""
        return {f"{name} mm": p["width (mm)"] for name, p in self.profiles.items()}


class QuartDVT(CatPhanBase):
    """A class for loading and analyzing CT DICOM files of a Quart phantom that comes with the Halcyon.
    Analyzes: HU Uniformity, Image Scaling & HU Linearity.
    """

    _demo_url = "quart.zip"
    _model = "Quart DVT"
    hu_origin_slice_variance = 300
    catphan_radius_mm = 80
    hu_module: QuartHUModule
    uniformity_module: QuartUniformityModule
    geometry_module: QuartGeometryModule

    @staticmethod
    def run_demo(show: bool = True):
        """Run the Quart algorithm with a head dataset."""
        quart = QuartDVT.from_demo_images()
        quart.analyze()
        print(quart.results())
        quart.plot_analyzed_image(show)

    def analyze(
        self,
        hu_tolerance: Union[int, float] = 40,
        scaling_tolerance: Union[int, float] = 1,
        thickness_tolerance: Union[int, float] = 0.2,
        cnr_threshold: Union[int, float] = 5,
    ):
        self.hu_module = QuartHUModule(
            self,
            offset=0,
            hu_tolerance=hu_tolerance,
            thickness_tolerance=thickness_tolerance,
            scaling_tolerance=scaling_tolerance,
        )
        self.uniformity_module = QuartUniformityModule(
            self, offset=UNIFORMITY_OFFSET_MM, tolerance=hu_tolerance
        )
        self.geometry_module = QuartGeometryModule(
            self, tolerance=3, offset=GEOMETRY_OFFSET_MM
        )

    def plot_analyzed_image(self, show: bool = True, **plt_kwargs) -> None:
        """Plot the images used in the calculation and summary data.

        Parameters
        ----------
        show : bool
            Whether to plot the image or not.
        plt_kwargs : dict
            Keyword args passed to the plt.figure() method. Allows one to set things like figure size.
        """
        # set up grid and axes
        plt.figure(**plt_kwargs)
        grid_size = (2, 3)
        hu_ax = plt.subplot2grid(grid_size, (0, 1))
        self.hu_module.plot(hu_ax)
        hu_lin_ax = plt.subplot2grid(grid_size, (0, 2))
        self.hu_module.plot_linearity(hu_lin_ax)
        unif_ax = plt.subplot2grid(grid_size, (1, 0))
        self.uniformity_module.plot(unif_ax)
        unif_prof_ax = plt.subplot2grid(grid_size, (1, 1), colspan=2)
        self.uniformity_module.plot_profiles(unif_prof_ax)
        geometry_ax = plt.subplot2grid(grid_size, (0, 0))
        self.geometry_module.plot(geometry_ax)

        # finish up
        plt.tight_layout()
        if show:
            plt.show()

    def plot_analyzed_subimage(self, *args, **kwargs) -> None:
        raise NotImplementedError()

    def results(self, as_str: bool = True) -> Union[str, Tuple[str, ...]]:
        """Return the results of the analysis as a string. Use with print()."""
        items = (
            f"\n - {self._model} QA Test - \n",
            f"HU Linearity ROIs: {self.hu_module.roi_vals_as_str}\n",
            f"HU Passed?: {self.hu_module.passed_hu}\n",
            f"Measured Slice Thickness (mm): {self.hu_module.meas_slice_thickness:2.3f}\n",
            f"Slice Thickness Passed? {self.hu_module.passed_thickness}\n",
            f"Uniformity ROIs: {self.uniformity_module.roi_vals_as_str}\n",
            f"Uniformity Passed?: {self.uniformity_module.overall_passed}\n",
            f"Geometric width: {self.geometry_module.distances()}",
        )
        if as_str:
            return "\n".join(items)
        else:
            return items

    def results_data(self, as_dict: bool = False) -> Union[QuartDVTResult, dict]:
        """Return results in a data structure for more programmatic use."""
        data = QuartDVTResult(
            phantom_model=self._model,
            phantom_roll_deg=self.catphan_roll,
            origin_slice=self.origin_slice,
            num_images=self.num_images,
            uniformity_module=QuartUniformityModuleOutput(
                offset=UNIFORMITY_OFFSET_MM,
                roi_settings=self.uniformity_module.roi_settings,
                rois=rois_to_results(self.uniformity_module.rois),
                passed=self.uniformity_module.overall_passed,
            ),
            geometric_module=QuartGeometryModuleOutput(
                offset=GEOMETRY_OFFSET_MM,
                roi_settings=self.geometry_module.roi_settings,
                rois=rois_to_results(self.geometry_module.rois),
                distances=self.geometry_module.distances(),
            ),
            hu_module=QuartHUModuleOutput(
                offset=0,
                roi_settings=self.hu_module.roi_settings,
                rois=rois_to_results(self.hu_module.rois),
                measured_slice_thickness_mm=self.hu_module.meas_slice_thickness,
                signal_to_noise=self.hu_module.signal_to_noise,
                contrast_to_noise=self.hu_module.contrast_to_noise,
            ),
        )
        if as_dict:
            return dataclasses.asdict(data)
        else:
            return data

    def plot_images(self, show: bool = True, **plt_kwargs) -> Dict[str, plt.Figure]:
        """Plot all the individual images separately.

        Parameters
        ----------
        show
            Whether to show the images.
        plt_kwargs
            Keywords to pass to matplotlib for figure customization.
        """
        figs = {}
        # plot the images
        modules = {
            "HU linearity": self.hu_module,
            "HU uniformity": self.uniformity_module,
            "Geometry": self.geometry_module,
        }
        for key, module in modules.items():
            fig, ax = plt.subplots(**plt_kwargs)
            module.plot(ax)
            figs[key] = fig

        if show:
            plt.show()
        return figs

    def save_images(
        self,
        directory: Optional[Union[Path, str]] = None,
        to_stream: bool = False,
        **plt_kwargs,
    ) -> Union[List[Path], Dict[str, BytesIO]]:
        """Save separate images to disk or stream.

        Parameters
        ----------
        directory
            The directory to write the images to. If None, will use current working directory
        to_stream
            Whether to write to stream or disk. If True, will return streams. Directory is ignored in that scenario.
        plt_kwargs
            Keywords to pass to matplotlib for figure customization.
        """
        figs = self.plot_images(show=False, **plt_kwargs)
        paths = []
        streams = {}
        for name, fig in figs.items():
            if to_stream:
                path = io.BytesIO()
            else:
                destination = Path(directory) or Path.cwd()
                path = (destination / name).with_suffix(".png").absolute()
            fig.savefig(path)
            paths.append(path)
            streams[name] = path
        if to_stream:
            return streams
        else:
            return paths

    def publish_pdf(
        self,
        filename: Union[str, Path],
        notes: Optional[str] = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ) -> None:
        """Publish (print) a PDF containing the analysis and quantitative results.

        Parameters
        ----------
        filename : (str, file-like object}
            The file to write the results to.
        notes : str, list of strings
            Text; if str, prints single line.
            If list of strings, each list item is printed on its own line.
        open_file : bool
            Whether to open the file using the default program after creation.
        metadata : dict
            Extra data to be passed and shown in the PDF. The key and value will be shown with a colon.
            E.g. passing {'Author': 'James', 'Unit': 'TrueBeam'} would result in text in the PDF like:
            --------------
            Author: James
            Unit: TrueBeam
            --------------
        """
        analysis_title = f"{self._model} Analysis"
        analysis_images = self.save_images(to_stream=True)

        canvas = pdf.PylinacCanvas(
            filename, page_title=analysis_title, metadata=metadata
        )
        if notes is not None:
            canvas.add_text(text="Notes:", location=(1, 4.5), font_size=14)
            canvas.add_text(text=notes, location=(1, 4))

        shortened_texts = [
            textwrap.wrap(r, width=110) for r in self.results(as_str=False)
        ]
        idx = 0
        for items in enumerate(shortened_texts):
            for text in items:
                canvas.add_text(text=text, location=(1.5, 25 - idx * 0.5))
                idx += 1
        for page, img in enumerate(analysis_images.values()):
            canvas.add_new_page()
            canvas.add_image(img, location=(1, 5), dimensions=(18, 18))
        canvas.finish()

        if open_file:
            webbrowser.open(filename)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pylinac',
 'pylinac.calibration',
 'pylinac.core',
 'pylinac.core.image_generator']

package_data = \
{'': ['*'], 'pylinac': ['files/*']}

install_requires = \
['Pillow>=4.0',
 'argue>=0.3.1,<0.4.0',
 'cached-property>=1.5.2,<1.6.0',
 'matplotlib>=2.0',
 'numpy>=0.16',
 'py-linq>=1.3.0,<2.0.0',
 'pydicom>=2.0',
 'reportlab>=3.3',
 'scikit-image>=0.17.3',
 'scipy>=1.1',
 'tqdm>=3.8']

setup_kwargs = {
    'name': 'pylinac',
    'version': '3.3.0',
    'description': 'A toolkit for performing TG-142 QA-related tasks on a linear accelerator',
    'long_description': 'Pylinac\n=======\n\n.. image:: https://storage.googleapis.com/pylinac_demo_files/Pylinac_Full_cropped.png\n    :width: 100%\n    :target: https://github.com/jrkerns/pylinac\n    :align: center\n\n|\n\n.. image:: https://img.shields.io/pypi/v/pylinac.svg\n    :target: https://pypi.python.org/pypi/pylinac\n    :alt: Latest Version\n\n.. image:: https://img.shields.io/pypi/l/pylinac.svg\n    :target: https://choosealicense.com/licenses/mit/\n    :alt: MIT\n\n.. image:: https://travis-ci.org/jrkerns/pylinac.svg?branch=master\n    :target: https://travis-ci.org/jrkerns/pylinac\n    :alt: Travis CI\n\n.. image:: https://coveralls.io/repos/github/jrkerns/pylinac/badge.svg?branch=master\n    :target: https://coveralls.io/github/jrkerns/pylinac?branch=master\n    :alt: Code Coverage\n\n.. image:: https://readthedocs.org/projects/pylinac/badge/?version=stable\n    :target: http://pylinac.readthedocs.org/en/stable/?badge=stable\n    :alt: Documentation Status\n\n\nPylinac provides TG-142 quality assurance (QA) tools to Python programmers as well as non-programmers in the field of \ntherapy medical physics.\n\nPylinac contains high-level modules for automatically analyzing images and data generated by linear accelerators.\nMost scripts can be utilized with <10 lines of code. The built-in GUI is also straight-forward although programming experience may help identify errors.\n\nThe library also contains lower-level `hackable modules & tools <http://pylinac.readthedocs.org/en/stable/pylinac_core_hacking.html>`_\nfor creating your own image analysis algorithms.\n\nThe major features of the entire package include:\n\n* PDF report generation for solid documentation\n* Automatic phantom registration even if you don\'t set up your phantom perfect\n* Image loading from file, ZIP archives, or URLs\n\nDocumentation\n-------------\n\nTo get started, install the package, run the demos, view the API docs, and learn the module design, visit the\n`Full Documentation <http://pylinac.readthedocs.org/>`_ on Read The Docs.\n\nInstallation\n------------\n\nInstall via ``pip``:\n\n.. code-block:: bash\n\n    $ pip install pylinac\n\nSee the `Installation page <http://pylinac.readthedocs.io/en/stable/installation.html>`_ for further details.\n\nTools/Quick Start Guide\n-----------------------\n\nBelow are the high-level tools currently available:\n\n* `TG-51 & TRS-398 Absolute Dose Calibration <http://pylinac.readthedocs.org/en/stable/calibration_docs.html>`_ -\n   Input the raw data and pylinac can calculate either individual values (kQ, PDDx, Pion, etc) or use the\n   provided classes to input all measurement data and have it calculate all factors and dose values automatically.\n\n   Example script:\n\n   .. code-block:: python\n\n            from pylinac import tg51, trs398\n\n            ENERGY = 6\n            TEMP = 22.1\n            PRESS = tg51.mmHg2kPa(755.0)\n            CHAMBER = \'30013\'  # PTW\n            P_ELEC = 1.000\n            ND_w = 5.443  # Gy/nC\n            MU = 200\n            CLINICAL_PDD = 66.5\n\n            tg51_6x = tg51.TG51Photon(\n                unit=\'TrueBeam1\',\n                chamber=CHAMBER,\n                temp=TEMP, press=PRESS,\n                n_dw=ND_w, p_elec=P_ELEC,\n                measured_pdd10=66.4, lead_foil=None,\n                clinical_pdd10=66.5, energy=ENERGY,\n                voltage_reference=-300, voltage_reduced=-150,\n                m_reference=(25.65, 25.66, 25.65),\n                m_opposite=(25.64, 25.65, 25.65),\n                m_reduced=(25.64, 25.63, 25.63),\n                mu=MU, tissue_correction=1.0\n            )\n\n            # Done!\n            print(tg51_6x.dose_mu_dmax)\n\n            # examine other parameters\n            print(tg51_6x.pddx)\n            print(tg51_6x.kq)\n            print(tg51_6x.p_ion)\n\n            # change readings if you adjust output\n            tg51_6x.m_reference_adjusted = (25.44, 25.44, 25.43)\n            # print new dose value\n            print(tg51_6x.dose_mu_dmax_adjusted)\n\n            # generate a PDF for record-keeping\n            tg51_6x.publish_pdf(\'TB1 6MV TG-51.pdf\', notes=[\'My notes\', \'I used Pylinac to do this; so easy!\'], open_file=False)\n\n            # TRS-398 is very similar and just as easy!\n\n* `Planar Phantom Analysis (Leeds TOR, StandardImaging QC-3 & QC-kV, Las Vegas, Doselab MC2 (kV & MV), SNC kV & MV, PTW EPID QC) <http://pylinac.readthedocs.org/en/stable/planar_imaging.html>`_ -\n\n   Features:\n\n   * **Automatic phantom localization** - Set up your phantom any way you like; automatic positioning,\n     angle, and inversion correction mean you can set up how you like, nor will setup variations give you headache.\n   * **High and low contrast determination** - Analyze both low and high contrast ROIs. Set thresholds\n     as you see fit.\n\n   Example script:\n\n   .. code-block:: python\n\n        from pylinac import LeedsTOR, StandardImagingQC3, LasVegas, DoselabMC2kV, DoselabMC2MV\n\n        leeds = LeedsTOR("my_leeds.dcm")\n        leeds.analyze()\n        leeds.plot_analyzed_image()\n        leeds.publish_pdf()\n\n        qc3 = StandardImagingQC3("my_qc3.dcm")\n        qc3.analyze()\n        qc3.plot_analyzed_image()\n        qc3.publish_pdf(\'qc3.pdf\')\n\n        lv = LasVegas("my_lv.dcm")\n        lv.analyze()\n        lv.plot_analyzed_image()\n        lv.publish_pdf(\'lv.pdf\', open_file=True)  # open the PDF after publishing\n\n        ...\n\n* `Winston-Lutz Analysis <http://pylinac.readthedocs.org/en/stable/winston_lutz.html>`_ -\n    The Winston-Lutz module analyzes EPID images taken of a small radiation field and BB to determine the 2D\n    distance from BB to field CAX. Additionally, the isocenter size of the gantry, collimator, and couch can\n    all be determined *without the BB being at isocenter*. Analysis is based on\n    `Winkler et al <http://iopscience.iop.org/article/10.1088/0031-9155/48/9/303/meta;jsessionid=269700F201744D2EAB897C14D1F4E7B3.c2.iopscience.cld.iop.org>`_\n    , `Du et al <http://scitation.aip.org/content/aapm/journal/medphys/37/5/10.1118/1.3397452>`_, and\n    `Low et al <https://aapm.onlinelibrary.wiley.com/doi/abs/10.1118/1.597475>`_.\n\n    Features:\n\n    * **Couch shift instructions** - After running a WL test, get immediate feedback on how to shift the couch.\n      Couch values can also be passed in and the new couch values will be presented so you don\'t have to do that pesky conversion.\n      "Do I subtract that number or add it?"\n    * **Automatic field & BB positioning** - When an image or directory is loaded, the field CAX and the BB\n      are automatically found, along with the vector and scalar distance between them.\n    * **Isocenter size determination** - Using backprojections of the EPID images, the 3D gantry isocenter size\n      and position can be determined *independent of the BB position*. Additionally, the 2D planar isocenter size\n      of the collimator and couch can also be determined.\n    * **Image plotting** - WL images can be plotted separately or together, each of which shows the field CAX, BB and\n      scalar distance from BB to CAX.\n    * **Axis deviation plots** - Plot the variation of the gantry, collimator, couch, and EPID in each plane\n      as well as RMS variation.\n    * **File name interpretation** - Rename DICOM filenames to include axis information for linacs that don\'t include\n      such information in the DICOM tags. E.g. "myWL_gantry45_coll0_couch315.dcm".\n\n    Example script:\n\n    .. code-block:: python\n\n        from pylinac import WinstonLutz\n\n        wl = WinstonLutz("wl/image/directory")  # images are analyzed upon loading\n        wl.plot_summary()\n        print(wl.results())\n        wl.publish_pdf(\'my_wl.pdf\')\n\n* `Starshot Analysis <http://pylinac.readthedocs.org/en/stable/starshot_docs.html>`_ -\n    The Starshot module analyses a starshot image made of radiation spokes, whether gantry, collimator, MLC or couch.\n    It is based on ideas from `Depuydt et al <http://iopscience.iop.org/0031-9155/57/10/2997>`_\n    and `Gonzalez et al <http://dx.doi.org/10.1118/1.1755491>`_.\n\n    Features:\n\n    * **Analyze scanned film images, single EPID images, or a set of EPID images** -\n      Any image that you can load in can be analyzed, including 1 or a set of EPID DICOM images and\n      films that have been digitally scanned.\n    * **Any image size** - Have machines with different EPIDs? Scanned your film at different resolutions? No problem.\n    * **Dose/OD can be inverted** - Whether your device/image views dose as an increase in value or a decrease, pylinac\n      will detect it and invert if necessary.\n    * **Automatic noise detection & correction** - Sometimes there\'s dirt on the scanned film; sometimes there\'s a dead pixel on the EPID.\n      Pylinac will detect these spurious noise signals and can avoid or account for them.\n    * **Accurate, FWHM star line detection** - Pylinac uses not simply the maximum value to find the center of a star line,\n      but analyzes the entire star profile to determine the center of the FWHM, ensuring small noise or maximum value bias is avoided.\n    * **Adaptive searching** - If you passed pylinac a set of parameters and a good result wasn\'t found, pylinac can recover and\n      do an adaptive search by adjusting parameters to find a "reasonable" wobble.\n\n    Example script:\n\n    .. code-block:: python\n\n        from pylinac import Starshot\n\n        star = Starshot("mystarshot.tif")\n        star.analyze(radius=0.75, tolerance=1.0, fwhm=True)\n        print(star.results())  # prints out wobble information\n        star.plot_analyzed_image()  # shows a matplotlib figure\n        star.publish_pdf()  # publish a PDF report\n\n* `VMAT QA <http://pylinac.readthedocs.org/en/stable/vmat_docs.html>`_ -\n    The VMAT module consists of two classes: ``DRGS`` and ``DRMLC``, which are capable of loading an EPID DICOM Open field image and MLC field image and analyzing the\n    images according to the Varian RapidArc QA tests and procedures, specifically the Dose-Rate & Gantry-Speed (DRGS) and MLC speed (MLCS) tests.\n\n    Features:\n\n    * **Do both tests** - Pylinac can handle either DRGS or DRMLC tests.\n    * **Adjust for offsets** - Older VMAT patterns were off-center. Pylinac will find the field regardless.\n\n    Example script:\n\n    .. code-block:: python\n\n        from pylinac import DRGS, DRMLC\n\n        drgs = DRGS(image_paths=["path/to/DRGSopen.dcm", "path/to/DRGSdmlc.dcm"])\n        drgs.analyze(tolerance=1.5)\n        print(drgs.results())  # prints out ROI information\n        drgs.plot_analyzed_image()  # shows a matplotlib figure\n        drgs.publish_pdf(\'mydrgs.pdf\')  # generate a PDF report\n\n* `CatPhan, Quart, ACR phantom QA <http://pylinac.readthedocs.org/en/stable/cbct_docs.html>`_ -\n    The CBCT module automatically analyzes DICOM images of a CatPhan 504, 503, 600, 604, Quart DVT, and ACR CT/MR acquired when doing CT, CBCT, or MR quality assurance. It can load a folder or zip file that\n    the images are in and automatically correct for phantom setup in 6 axes.\n    CatPhans analyze the HU regions and image scaling (CTP404), the high-contrast line pairs (CTP528) to calculate the modulation transfer function (MTF), and the HU\n    uniformity (CTP486) on the corresponding slice. Quart and ACR analyze similar metrics where possible.\n\n    Features:\n\n    * **Automatic phantom registration** - Your phantom can be tilted, rotated, or translated--pylinac will register the phantom.\n    * **Automatic testing of all major modules** - Major modules are automatically registered and analyzed.\n    * **Any scan protocol** - Scan your CatPhan with any protocol; or even scan it in a regular CT scanner.\n      Any field size or field extent is allowed.\n    * **Customize modules** - You can easily override settings in the event you have a custom scenario such as a partial scan.\n\n    Example script:\n\n    .. code-block:: python\n\n        from pylinac import CatPhan504, CatPhan503, CatPhan600, CatPhan604, QuartDVT, ACRCT, ACRMRILarge\n\n        # for this example, we\'ll use the CatPhan504\n        cbct = CatPhan504("my/cbct_image_folder")\n        cbct.analyze(hu_tolerance=40, scaling_tolerance=1, thickness_tolerance=0.2, low_contrast_threshold=1)\n        print(cbct.results())\n        cbct.plot_analyzed_image()\n        cbct.publish_pdf(\'mycbct.pdf\')\n\n* `Log Analysis <http://pylinac.readthedocs.org/en/stable/log_analyzer.html>`_ -\n    The log analyzer module reads and parses Varian linear accelerator machine logs, both Dynalogs and Trajectory logs. The module also\n    calculates actual and expected fluences as well as performing gamma evaluations. Data is structured to be easily accessible and\n    easily plottable.\n\n    Unlike most other modules of pylinac, the log analyzer module has no end goal. Data is parsed from the logs, but what is done with that\n    info, and which info is analyzed is up to the user.\n\n    Features:\n\n    * **Analyze Dynalogs or Trajectory logs** - Either platform is supported. Tlog versions 2.1 and 3.0 supported.\n    * **Save Trajectory log data to CSV** - The Trajectory log binary data format does not allow for easy export of data. Pylinac lets you do\n      that so you can use Excel or other software that you use with Dynalogs.\n    * **Plot or analyze any axis** - Every data axis can be plotted: the actual, expected, and even the difference.\n    * **View actual or expected fluences & calculate gamma** - View fluences and gamma maps for any log.\n    * **Anonymization** - Anonymize your logs so you can share them with others.\n\n    Example script:\n\n    .. code-block:: python\n\n        from pylinac import load_log\n\n        tlog = load_log("tlog.bin")\n        # after loading, explore any Axis of the Varian structure\n        tlog.axis_data.gantry.plot_actual()  # plot the gantry position throughout treatment\n        tlog.fluence.gamma.calc_map(doseTA=1, distTA=1, threshold=10, resolution=0.1)\n        tlog.fluence.gamma.plot_map()  # show the gamma map as a matplotlib figure\n        tlog.publish_pdf()  # publish a PDF report\n\n        dlog = load_log("dynalog.dlg")\n        ...\n\n* `Picket Fence MLC Analysis <http://pylinac.readthedocs.org/en/stable/picketfence.html>`_ -\n    The picket fence module is meant for analyzing EPID images where a "picket fence" MLC pattern has been made.\n    Physicists regularly check MLC positioning through this test. This test can be done using film and one can\n    "eyeball" it, but this is the 21st century and we have numerous ways of quantifying such data. This module\n    attains to be one of them. It will load in an EPID dicom image and determine the MLC peaks, error of each MLC\n    pair to the picket, and give a few visual indicators for passing/warning/failing.\n\n    Features:\n\n    * **Preset & customizable MLC configurations** - Standard configurations are built-in and you can create your own configuration of leaves if needed.\n    * **Easy-to-read pass/warn/fail overlay** - Analysis gives you easy-to-read tools for determining the status of an MLC pair.\n    * **Any Source-to-Image distance** - Whatever your clinic uses as the SID for picket fence, pylinac can account for it.\n    * **Account for panel translation** - Have an off-CAX setup? No problem. Translate your EPID and pylinac knows.\n    * **Account for panel sag** - If your EPID sags at certain angles, just tell pylinac and the results will be shifted.\n\n    Example script:\n\n    .. code-block:: python\n\n        from pylinac import PicketFence\n\n        pf = PicketFence("mypf.dcm")\n        pf.analyze(tolerance=0.5, action_tolerance=0.25)\n        print(pf.results())\n        pf.plot_analyzed_image()\n        pf.publish_pdf()\n\n* `Open Field Analysis <http://pylinac.readthedocs.org/en/stable/field_analysis.html>`_ -\n    Field analysis from a digital image such as EPID DICOM or 2D device array can easily be analyzed. The module contains built-in\n    flatness and symmetry equation definitions but is extensible to quickly create custom F&S equations.\n\n    Features:\n    * **EPID or device data** - Any EPID image or the SNC Profiler.\n    * **Built-in F&S equations** - The common Elekta, Varian, and Siemens definitions are included\n    * **Extensible equations** - Adding custom equations for image metrics are easy\n\n    Example script:\n\n    .. code-block:: python\n\n        from pylinac import FieldAnalysis, DeviceFieldAnalysis, Protocol\n\n        fa = FieldAnalysis(path="myFS.dcm")  # equivalently, DeviceFieldAnalysis\n        fa.analyze(protocol=Protocol.VARIAN)\n        # print results\n        print(fa.results())\n        # get results as a dict\n        fa.results_data()\n        # plot results\n        fa.plot_analyzed_image()\n        # publish a PDF file\n        fa.publish_pdf(filename=\'my field analysis.pdf\')\n\nDiscussion\n----------\n\nHave questions? Ask them on the `pylinac discussion forum <https://groups.google.com/forum/#!forum/pylinac>`_.\n\nContributing\n------------\n\nContributions to pylinac can be many. The most useful things a non-programmer can contribute are images to analyze and bug reports. If\nyou have VMAT images, starshot images, machine log files, CBCT DICOM files, or anything else you want analyzed, upload them privately\n`here <https://forms.gle/RBR5ubFvjogE9iC67>`_.\n',
    'author': 'James Kerns',
    'author_email': 'jkerns100@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jrkerns/pylinac',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.10,<4.0.0',
}


setup(**setup_kwargs)



import webbrowser
from abc import ABC
from datetime import datetime
from typing import Union, Optional

import argue
import numpy as np

from . import tg51 as _tg51
from .tg51 import (
    mmHg2kPa,
    mbar2kPa,
    fahrenheit2celsius,
    tpr2010_from_pdd2010,
    MIN_PTP,
    MAX_PTP,
    MIN_PELEC,
    MAX_PELEC,
    MIN_PPOL,
    MAX_PPOL,
    MIN_PION,
    MAX_PION,
)  # make available to module
from ..core.pdf import PylinacCanvas
from ..core.typing import NumberOrArray
from ..core.utilities import is_close, Structure

V1_V2_FITS = {
    2.0: {"a0": 2.337, "a1": -3.636, "a2": 2.299},
    2.5: {"a0": 1.474, "a1": -1.587, "a2": 1.114},
    3.0: {"a0": 1.198, "a1": -0.875, "a2": 0.677},
    3.5: {"a0": 1.080, "a1": -0.542, "a2": 0.463},
    4.0: {"a0": 1.022, "a1": -0.363, "a2": 0.341},
    5.0: {"a0": 0.975, "a1": -0.188, "a2": 0.214},
}

KQ_PHOTON_TPRS = (
    0.50,
    0.53,
    0.56,
    0.59,
    0.62,
    0.65,
    0.68,
    0.70,
    0.72,
    0.74,
    0.76,
    0.78,
    0.80,
    0.82,
    0.84,
)

# chamber kQ values from table 6.III
KQ_PHOTON_CHAMBERS = {
    # Capintec
    "PR-06C/G": (
        1.001,
        1.001,
        1.000,
        0.998,
        0.998,
        0.995,
        0.992,
        0.990,
        0.988,
        0.984,
        0.980,
        0.972,
        0.965,
        0.956,
        0.944,
    ),
    # Exradin
    "A12": (
        1.001,
        1.001,
        1.000,
        1.000,
        0.999,
        0.997,
        0.994,
        0.992,
        0.990,
        0.986,
        0.981,
        0.974,
        0.966,
        0.957,
        0.944,
    ),
    # Nuclear Associates
    "30-751": (
        1.002,
        1.002,
        1.000,
        0.999,
        0.997,
        0.994,
        0.991,
        0.989,
        0.985,
        0.981,
        0.977,
        0.969,
        0.961,
        0.953,
        0.940,
    ),
    "30-752": (
        1.004,
        1.003,
        1.001,
        1.000,
        0.998,
        0.996,
        0.993,
        0.991,
        0.989,
        0.985,
        0.981,
        0.974,
        0.967,
        0.959,
        0.947,
    ),
    # NE
    "2505": (
        1.001,
        1.001,
        1.000,
        0.999,
        0.997,
        0.994,
        0.991,
        0.988,
        0.984,
        0.980,
        0.975,
        0.967,
        0.959,
        0.950,
        0.937,
    ),
    "2505/A": (
        1.005,
        1.003,
        1.001,
        0.997,
        0.995,
        0.990,
        0.985,
        0.982,
        0.978,
        0.974,
        0.969,
        0.962,
        0.955,
        0.947,
        0.936,
    ),
    "2505/3, 3A": (
        1.005,
        1.004,
        1.002,
        1.000,
        0.998,
        0.995,
        0.993,
        0.991,
        0.989,
        0.986,
        0.982,
        0.975,
        0.969,
        0.961,
        0.949,
    ),
    "2505/3, 3B": (
        1.006,
        1.004,
        1.001,
        0.999,
        0.996,
        0.991,
        0.987,
        0.984,
        0.980,
        0.976,
        0.971,
        0.964,
        0.957,
        0.950,
        0.938,
    ),
    "2571": (
        1.005,
        1.004,
        1.002,
        1.000,
        0.998,
        0.995,
        0.993,
        0.991,
        0.989,
        0.986,
        0.982,
        0.975,
        0.969,
        0.961,
        0.949,
    ),
    "2581": (
        1.005,
        1.003,
        1.001,
        0.998,
        0.995,
        0.991,
        0.986,
        0.983,
        0.980,
        0.975,
        0.970,
        0.963,
        0.956,
        0.949,
        0.937,
    ),
    # PTW
    "30001": (
        1.004,
        1.003,
        1.001,
        0.999,
        0.997,
        0.994,
        0.990,
        0.988,
        0.985,
        0.981,
        0.976,
        0.969,
        0.962,
        0.955,
        0.943,
    ),
    "30010": (
        1.004,
        1.003,
        1.001,
        0.999,
        0.997,
        0.994,
        0.990,
        0.988,
        0.985,
        0.981,
        0.976,
        0.969,
        0.962,
        0.955,
        0.943,
    ),
    "30002": (
        1.006,
        1.004,
        1.001,
        0.999,
        0.997,
        0.994,
        0.992,
        0.990,
        0.987,
        0.984,
        0.980,
        0.973,
        0.967,
        0.959,
        0.948,
    ),
    "30011": (
        1.006,
        1.004,
        1.001,
        0.999,
        0.997,
        0.994,
        0.992,
        0.990,
        0.987,
        0.984,
        0.980,
        0.973,
        0.967,
        0.959,
        0.948,
    ),
    "30004": (
        1.006,
        1.005,
        1.002,
        1.000,
        0.999,
        0.996,
        0.994,
        0.992,
        0.989,
        0.986,
        0.982,
        0.976,
        0.969,
        0.962,
        0.950,
    ),
    "30012": (
        1.006,
        1.005,
        1.002,
        1.000,
        0.999,
        0.996,
        0.994,
        0.992,
        0.989,
        0.986,
        0.982,
        0.976,
        0.969,
        0.962,
        0.950,
    ),
    "30006": (
        1.002,
        1.002,
        1.000,
        0.999,
        0.997,
        0.994,
        0.990,
        0.988,
        0.984,
        0.980,
        0.975,
        0.968,
        0.960,
        0.952,
        0.940,
    ),
    "30013": (
        1.002,
        1.002,
        1.000,
        0.999,
        0.997,
        0.994,
        0.990,
        0.988,
        0.984,
        0.980,
        0.975,
        0.968,
        0.960,
        0.952,
        0.940,
    ),
}

KQ_ELECTRON_R50S = (4.0, 4.5, 5.0, 5.5, 6.0, 7.0, 8.0, 10.0, 13.0, 16.0, 20.0)

KQ_ELECTRON_CHAMBERS = {
    # Capintec
    "PR06C": (
        0.916,
        0.914,
        0.912,
        0.911,
        0.909,
        0.906,
        0.904,
        0.899,
        0.891,
        0.884,
        0.874,
    ),
    # Exradin
    "A12": (
        0.921,
        0.919,
        0.918,
        0.916,
        0.914,
        0.911,
        0.909,
        0.903,
        0.896,
        0.888,
        0.878,
    ),
    # NE
    "2571": (
        0.918,
        0.916,
        0.915,
        0.913,
        0.911,
        0.909,
        0.906,
        0.901,
        0.893,
        0.886,
        0.876,
    ),
    "2581": (
        0.899,
        0.898,
        0.896,
        0.894,
        0.893,
        0.890,
        0.888,
        0.882,
        0.875,
        0.868,
        0.859,
    ),
    # PTW
    "30001": (
        0.911,
        0.909,
        0.907,
        0.905,
        0.904,
        0.901,
        0.898,
        0.893,
        0.885,
        0.877,
        0.868,
    ),
    "30010": (
        0.911,
        0.909,
        0.907,
        0.905,
        0.904,
        0.901,
        0.898,
        0.893,
        0.885,
        0.877,
        0.868,
    ),
    "30002": (
        0.916,
        0.914,
        0.912,
        0.910,
        0.909,
        0.906,
        0.903,
        0.897,
        0.890,
        0.882,
        0.873,
    ),
    "30011": (
        0.916,
        0.914,
        0.912,
        0.910,
        0.909,
        0.906,
        0.903,
        0.897,
        0.890,
        0.882,
        0.873,
    ),
    "30004": (
        0.920,
        0.918,
        0.916,
        0.915,
        0.913,
        0.910,
        0.907,
        0.902,
        0.894,
        0.887,
        0.877,
    ),
    "30012": (
        0.920,
        0.918,
        0.916,
        0.915,
        0.913,
        0.910,
        0.907,
        0.902,
        0.894,
        0.887,
        0.877,
    ),
    "30006": (
        0.911,
        0.909,
        0.907,
        0.906,
        0.904,
        0.901,
        0.898,
        0.893,
        0.885,
        0.878,
        0.868,
    ),
    "30013": (
        0.911,
        0.909,
        0.907,
        0.906,
        0.904,
        0.901,
        0.898,
        0.893,
        0.885,
        0.878,
        0.868,
    ),
    # Wellhofer
    "FC65-P": (
        0.914,
        0.912,
        0.911,
        0.909,
        0.907,
        0.904,
        0.902,
        0.896,
        0.889,
        0.881,
        0.872,
    ),
    "FC65-G": (
        0.920,
        0.918,
        0.916,
        0.914,
        0.913,
        0.910,
        0.907,
        0.902,
        0.894,
        0.887,
        0.877,
    ),
}

# Rename common functions from TG-51
k_tp = _tg51.p_tp
k_pol = _tg51.p_pol
z_ref = _tg51.d_ref
r_50 = _tg51.r_50


def k_s(
    *,
    voltage_reference: int,
    voltage_reduced: int,
    m_reference: NumberOrArray,
    m_reduced: NumberOrArray,
) -> float:
    """Calculate the ion recombination effect using readings at two voltages. The voltages should have
    a ratio of 2, 2.5, 3, 3.5, 4, or 5.

    Parameters
    ----------
    voltage_reference : int
        The voltage at which calibration will be performed (e.g. 300V)
    voltage_reduced : int
        The voltage which is lower than reference (e.g. 150V)
    m_reference : array, float
        The reading(s) at the reference voltage.
    m_reduced : array, float
        The reading(s) at the reduced voltage.

    Returns
    -------
    k_s : float
        The ion recombination factor.

    Raises
    ------
    ValueError
        If the voltage ratio is not valid.
    ValueError
        If the calculated ks value is outside the range (1.0, 1.05).
    """
    v_ratio = voltage_reference / voltage_reduced
    _verify_voltage_ratio_is_valid(v_ratio)
    a = V1_V2_FITS[v_ratio]
    m_ratio = np.mean(m_reference) / np.mean(m_reduced)
    argue.verify_bounds(
        m_ratio,
        bounds=(MIN_PION, MAX_PION),
        message="Ks is out of bounds. Verify inputs or check chamber",
    )
    return float(a["a0"] + a["a1"] * m_ratio + a["a2"] * (m_ratio ** 2))


def _verify_voltage_ratio_is_valid(voltage_ratio):
    """Helper function to verify the voltage ratio of high/low is one of a valid set as defined by TRS-398 Table 4.VII"""
    if not is_close(voltage_ratio, target=(2, 2.5, 3, 3.5, 4, 5), delta=0.001):
        raise ValueError(
            "voltage_reference and voltage_reduced are not a valid ratio. Valid ratios are: 2, 2.5, 3, 3.5, 4, 5"
        )


@argue.options(chamber=KQ_PHOTON_CHAMBERS.keys())
@argue.bounds(tpr=(KQ_PHOTON_TPRS[0], KQ_PHOTON_TPRS[-1]))
def kq_photon(*, chamber: str, tpr: float) -> float:
    """Calculate the kQ factor for a photon beam given the chamber model and TPR20/10 using Table 6.III.
    Linear interpolation is used between given TPR ratios.

    Parameters
    ----------
    chamber : str
        Allowable chambers are those listed in Table 6.III that are also Farmer-type (e.g. Exradin A14 Farmer).
    tpr : float
        The ratio of measured TPR(20cm) / TPR(10cm). Note that this can also be calculated from PDD. See :func:`~pylinac.calibration.tg51.tpr2010_from_pdd2010`.

    Returns
    -------
    kQ : float
        The calculated kQ given table Table 6.III

    Raises
    ------
    KeyError
        If the passed chamber is not within the acceptable list.
    ValueError
        If the TPR is not within the range defined by Table 6.III
    """
    return np.interp([tpr], KQ_PHOTON_TPRS, KQ_PHOTON_CHAMBERS[chamber])[0]


@argue.options(chamber=KQ_ELECTRON_CHAMBERS.keys())
@argue.bounds(r_50=(KQ_ELECTRON_R50S[0], KQ_ELECTRON_R50S[-1]))
def kq_electron(*, chamber: str, r_50: float) -> float:
    """Calculate the kQ factor for an electron beam given the chamber model and R50 using Table 7.III.
    Linear interpolation is used between given R50 values.

    Parameters
    ----------
    chamber : str
        The Farmer-type chambers listed in Table 7.III (e.g. PTW 30004/30012).
    r_50 : float
        The depth of R50 in cm in water.

    Returns
    -------
    kQ : float
        The calculated kQ from Table 7.III

    Raises
    ------
    KeyError
        If the passed chamber is not within the acceptable list.
    ValueError
        If the R50 is not within the range defined by Table 7.III
    """
    return np.interp([r_50], KQ_ELECTRON_R50S, KQ_ELECTRON_CHAMBERS[chamber])[0]


def m_corrected(*, m_reference, k_tp, k_elec, k_pol, k_s) -> float:
    """The fully corrected chamber reading.

    Parameters
    ----------
    m_reference : array, float
        The chamber reading(s) at the calibration position.
    k_tp : float
        Temperature/Pressure correction. See :func:`~pylinac.calibration.tg51.p_tp`.
    k_elec : float
        Electrometer correction; given by the calibration laboratory.
    k_pol : float
        Polarity correction. See :func:`~pylinac.calibration.tg51.p_pol`.
    k_s : float
        Ion recombination correction. See :func:`~pylinac.calibration.trs398.k_s`.

    Returns
    -------
    m : float
        The fully corrected chamber reading.
    """
    argue.verify_bounds(k_tp, bounds=(MIN_PTP, MAX_PTP))
    argue.verify_bounds(k_elec, bounds=(MIN_PELEC, MAX_PELEC))
    argue.verify_bounds(k_pol, bounds=(MIN_PPOL, MAX_PPOL))
    argue.verify_bounds(k_s, bounds=(MIN_PION, MAX_PION))
    return float(np.mean(m_reference) * k_tp * k_elec * k_pol * k_s)


class TRS398Base(ABC, Structure):
    @property
    def k_tp(self):
        """Temperature/Pressure correction"""
        return k_tp(temp=self.temp, press=self.press)

    @property
    def k_pol(self):
        """Polarity correction"""
        return k_pol(m_reference=self.m_reference, m_opposite=self.m_opposite)

    @property
    def k_s(self):
        """Ionization collection correction."""
        return k_s(
            voltage_reference=self.voltage_reference,
            voltage_reduced=self.voltage_reduced,
            m_reference=self.m_reference,
            m_reduced=self.m_reduced,
        )

    @property
    def m_corrected(self):
        """Corrected chamber reading."""
        return m_corrected(
            m_reference=self.m_reference,
            k_tp=self.k_tp,
            k_elec=self.k_elec,
            k_pol=self.k_pol,
            k_s=self.k_s,
        )

    @property
    def dose_mu_zref(self):
        """cGy/MU at a zref depth."""
        return self.tissue_correction * self.m_corrected * self.n_dw * self.kq / self.mu

    @property
    def m_corrected_adjusted(self):
        """Corrected chamber reading after adjusting the output."""
        # TODO: add check for m adjusted != None
        return m_corrected(
            m_reference=self.m_reference_adjusted,
            k_tp=self.k_tp,
            k_elec=self.k_elec,
            k_pol=self.k_pol,
            k_s=self.k_s,
        )

    @property
    def dose_mu_zref_adjusted(self):
        """The dose/mu at 10cm depth after adjustment."""
        return (
            self.tissue_correction
            * self.m_corrected_adjusted
            * self.n_dw
            * self.kq
            / self.mu
        )

    @property
    def output_was_adjusted(self):
        """Boolean specifiying if output was adjusted."""
        return self.m_reference_adjusted is not None


class TRS398Photon(TRS398Base):
    """Calculation of dose to water at zmax and zref from a high energy photon beam. Setup can be SSD or SAD.

    Parameters
    ----------
    setup : {'SSD', 'SAD'}
        The physical setup of the calibration.
    institution : str
        Institution name.
    physicist : str
        Physicist performing calibration.
    unit : str
        Unit name; e.g. TrueBeam1.
    measurement_date : str
        Date of measurement. E.g. 10/22/2018.
    chamber : str
        Farmer-type chamber model from Table 6.III.
    n_dw : float
        NDw of the chamber given by the calibration laboratory.
    mu : float, int
        The number of MU given per reading
    energy : int
        Nominal energy of the linac in MV; e.g. 6. Bookkeeping only.
    fff : bool
        Whether the beam is FFF or flat. Bookkeeping only.
    tpr2010 : float
        The value of TPR(20)/TPR(10). Can be derived from PDD; see :func:`~pylinac.calibration.tg51.tpr2010_from_pdd2010`.
    press : float
        The value of pressure in kPa. Can be converted from mmHg and mbar; see :func:`~pylinac.calibration.tg51.mmHg2kPa` and :func:`~pylinac.calibration.tg51.mbar2kPa`.
    temp : float
        The temperature in Celsius.
    voltage_reference : int
        The reference voltage; i.e. the voltage for the calibration reading (e.g. 300V).
    voltage_reduced : int
        The reduced voltage, usually a fraction of the reference voltage (e.g. 150V).
    m_reference : array, float
        The reading(s) of the chamber at reference voltage.
    m_reduced : array, float
        The reading(s) of the chamber at the reduced voltage.
    m_opposite : array, float
        The reading(s) of the chamber at the opposite voltage from reference. Sign of the reading does not matter.
    k_elec : float
        The electrometer correction value given by the calibration laboratory.
    clinical_pdd_zref : optional, float
        The PDD at the depth of calibration. Use the actual percentage (e.g. 66.7 not 0.667). If not supplied the clinical_tmr_zref value must be supplied.
    clinical_tmr_zref : optional, float
        The TMR at the depth of calibration. If not supplied the clinical_pdd_zref value must be supplied.
    tissue_correction : float
        The correction of calibration to a medium other than water. Default value is 1 which is water. E.g. use 0.99 if calibrating to muscle.
    """

    @argue.options(setup=("SSD", "SAD"))
    @argue.options(chamber=KQ_PHOTON_CHAMBERS.keys())
    @argue.bounds(tpr2010=(KQ_PHOTON_TPRS[0], KQ_PHOTON_TPRS[-1]))
    def __init__(
        self,
        *,
        institution: str = "",
        physicist: str = "",
        unit: str = "",
        measurement_date: str = "",
        electrometer: str = "",
        setup: str,
        chamber: str,
        n_dw: float,
        mu: int,
        tpr2010: float,
        energy: int,
        fff: bool,
        press: float,
        temp: float,
        voltage_reference: int,
        voltage_reduced: int,
        m_reference: Union[tuple, float],
        m_reduced: Union[tuple, float],
        m_opposite: Union[tuple, float],
        k_elec: float,
        clinical_pdd_zref: Optional[float] = None,
        clinical_tmr_zref: Optional[float] = None,
        tissue_correction: float = 1.0,
    ):
        super().__init__(
            chamber=chamber,
            tpr2010=tpr2010,
            press=press,
            temp=temp,
            voltage_reference=voltage_reference,
            voltage_reduced=voltage_reduced,
            m_reference=m_reference,
            m_reduced=m_reduced,
            m_opposite=m_opposite,
            k_elec=k_elec,
            clinical_pdd_zref=clinical_pdd_zref,
            clinical_tmr_zref=clinical_tmr_zref,
            n_dw=n_dw,
            setup=setup,
            mu=mu,
            tissue_correction=tissue_correction,
            fff=fff,
            energy=energy,
            institution=institution,
            physicist=physicist,
            unit=unit,
            measurement_date=measurement_date,
            electrometer=electrometer,
        )
        self.m_reference_adjusted = None

    @property
    def kq(self):
        """kQ of the chamber and TPR."""
        return kq_photon(chamber=self.chamber, tpr=self.tpr2010)

    @property
    def dose_mu_zmax(self):
        """cGy/MU at a depth of zmax."""
        if self.setup == "SSD":
            return (100 * self.dose_mu_zref) / self.clinical_pdd_zref
        else:
            return self.dose_mu_zref / self.clinical_tmr_zref

    @property
    def dose_mu_zmax_adjusted(self):
        """The dose/mu at dmax depth after adjustment."""
        if self.setup == "SSD":
            return (100 * self.dose_mu_zref_adjusted) / self.clinical_pdd_zref
        else:
            return self.dose_mu_zref_adjusted / self.clinical_tmr_zref

    def publish_pdf(
        self,
        filename: str,
        notes: Optional[list] = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ):
        """Publish (print) a PDF containing the analysis and quantitative results.

        Parameters
        ----------
        filename : str, file-like object
            The file to write the results to.
        notes : str, list
            Any notes to be added to the report. If a string, adds everything as one line.
            If a list, must be a list of strings; each string item will be a new line.
        open_file : bool
            Whether to open the file after creation. Will use the default PDF program.
        metadata : dict
            Any data that should be appended to every page of the report. This differs from notes in that
            metadata is at the top of every page while notes is at the bottom of the report.
        """
        was_adjusted = "Yes" if self.output_was_adjusted else "No"
        title = f"TRS-398 Photon Report - {self.energy} MV"
        if self.fff:
            title += " FFF"

        canvas = PylinacCanvas(filename, page_title=title, metadata=metadata)
        text = [
            "Site Data:",
            f"Institution: {self.institution}",
            f"Performed by: {self.physicist}",
            f"Measurement Date: {self.measurement_date}",
            f'Date of Report: {datetime.now().strftime("%A, %B %d, %Y")}',
            f"Unit: {self.unit}",
            f"Energy: {self.energy} MV {'FFF' if self.fff else ''}",
            "",
            "Instrumentation:",
            f"Chamber: {self.chamber}",
            f"Chamber Calibration Factor Ndw (cGy/nC): {self.n_dw:2.3f}",
            f"Electrometer: {self.electrometer}",
            f"Kelec: {self.k_elec:2.3f}",
            f"MU: {self.mu}",
            "",
            "Beam Quality:",
            f"TPR(20)/TPR(10): {self.tpr2010:2.3f}",
            f"Determined kQ: {self.kq:2.3f}",
            "",
            "Chamber Corrections/Measurements:",
            f"Temperature (\N{DEGREE SIGN}C): {self.temp:2.1f}",
            f"Pressure (kPa): {self.press:2.1f}",
            f"Mraw @ ({self.voltage_reference}V, Reference) (nC): {self.m_reference}",
            f"Mraw @ ({self.voltage_reduced}V, Reduced) (nC): {self.m_reduced}",
            f"Mraw @ ({-self.voltage_reference}V, Opposite) (nC): {self.m_opposite}",
            f"Ktp: {self.k_tp:2.3f}",
            f"Ks: {self.k_s:2.3f}",
            f"Kpol: {self.k_pol:2.3f}",
            "",
            "Dose Determination:",
            f"Fully corrected M (nC): {self.m_corrected:2.3f}",
            f"Tissue correction (e.g. muscle): {self.tissue_correction:2.3f}",
            f"Dose/MU @ zref depth (cGy): {self.dose_mu_zref:2.3f}",
        ]
        if self.setup == "SSD":
            text.append(
                f"Clinical PDD (%): {self.clinical_pdd_zref:2.2f}",
            )
        else:
            text.append(
                f"Clinical TMR (%): {self.clinical_tmr_zref:2.2f}",
            )
        text.append(f"Dose/MU @ zmax (cGy): {self.dose_mu_zmax:2.3f}")
        text.append("")
        text.append(f"Output Adjustment?: {was_adjusted}")
        if was_adjusted == "Yes":
            text.append(
                f"Adjusted Mraw @ reference voltage (nC): {self.m_reference_adjusted}"
            )
            text.append(
                f"Adjusted fully corrected M (nC): {self.m_corrected_adjusted:2.3f}"
            )
            text.append(
                f"Adjusted Dose/MU @ zref depth (cGy): {self.dose_mu_zref_adjusted:2.3f}"
            )
            text.append(
                f"Adjusted Dose/MU @ zmax (cGy): {self.dose_mu_zmax_adjusted:2.3f}"
            )
        canvas.add_text(text=text, location=(2, 25.5), font_size=12)
        if notes is not None:
            canvas.add_text(text="Notes:", location=(12, 6.5), font_size=14)
            canvas.add_text(text=notes, location=(12, 6))

        canvas.finish()

        if open_file:
            webbrowser.open(filename)


class TRS398Electron(TRS398Base):
    """Calculation of dose to water at zmax and zref from a high energy electron beam.

    Parameters
    ----------
    institution : str
        Institution name.
    physicist : str
        Physicist performing calibration.
    unit : str
        Unit name; e.g. TrueBeam1.
    measurement_date : str
        Date of measurement. E.g. 10/22/2018.
    chamber : str
        Farmer-type chamber model from Table 6.III.
    n_dw : float
        NDw of the chamber given by the calibration laboratory.
    mu : float, int
        The number of MU given per reading.
    i_50 : float
        The depth of ionization 50% in cm.
    press : float
        The value of pressure in kPa. Can be converted from mmHg and mbar; see :func:`~pylinac.calibration.tg51.mmHg2kPa` and :func:`~pylinac.calibration.tg51.mbar2kPa`.
    temp : float
        The temperature in Celsius.
    voltage_reference : int
        The reference voltage; i.e. the voltage for the calibration reading (e.g. 300V).
    voltage_reduced : int
        The reduced voltage, usually a fraction of the reference voltage (e.g. 150V).
    m_reference : array, float
        The reading(s) of the chamber at reference voltage.
    m_reduced : array, float
        The reading(s) of the chamber at the reduced voltage.
    m_opposite : array, float
        The reading(s) of the chamber at the opposite voltage from reference. Sign of the reading does not matter.
    k_elec : float
        The electrometer correction value given by the calibration laboratory.
    pdd_zref : optional, float
        The PDD at the depth of calibration. Use the actual percentage (e.g. 66.7 not 0.667). If not supplied the tmr_zref value should be supplied.
    tissue_correction : float
        The correction of calibration to a medium other than water. Default value is 1 which is water. E.g. use 0.99 if calibrating to muscle.
    """

    def __init__(
        self,
        *,
        institution: str = "",
        physicist: str = "",
        unit: str = "",
        measurement_date: str = "",
        electrometer: str = "",
        energy: str,
        cone: str,
        chamber: str,
        n_dw: float,
        mu: int,
        i_50: float,
        press: float,
        temp: float,
        voltage_reference: int,
        voltage_reduced: int,
        m_reference: tuple,
        m_reduced: tuple,
        m_opposite: tuple,
        k_elec: float,
        clinical_pdd_zref: float,
        tissue_correction: float = 1.0,
    ):
        super().__init__(
            chamber=chamber,
            i_50=i_50,
            press=press,
            temp=temp,
            energy=energy,
            institution=institution,
            voltage_reference=voltage_reference,
            voltage_reduced=voltage_reduced,
            m_reference=m_reference,
            m_reduced=m_reduced,
            m_opposite=m_opposite,
            k_elec=k_elec,
            clinical_pdd_zref=clinical_pdd_zref,
            n_dw=n_dw,
            mu=mu,
            tissue_correction=tissue_correction,
            physicist=physicist,
            unit=unit,
            measurement_date=measurement_date,
            electrometer=electrometer,
            cone=cone,
        )
        self.m_reference_adjusted = None

    @property
    def r_50(self) -> float:
        """The depth of R50 in cm, derived from I50."""
        return r_50(i_50=self.i_50)

    @property
    def zref(self) -> float:
        """Depth of the reference point."""
        return z_ref(i_50=self.i_50)

    @property
    def kq(self):
        """kQ given the chamber and R50."""
        return kq_electron(chamber=self.chamber, r_50=self.r_50)

    @property
    def dose_mu_zmax(self):
        """cGy/MU at a depth of zmax."""
        return (100 * self.dose_mu_zref) / self.clinical_pdd_zref

    @property
    def dose_mu_zmax_adjusted(self):
        """The dose/mu at dmax depth after adjustment."""
        return (100 * self.dose_mu_zref_adjusted) / self.clinical_pdd_zref

    def publish_pdf(
        self,
        filename: str,
        notes: Optional[list] = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ):
        """Publish (print) a PDF containing the analysis and quantitative results.

        Parameters
        ----------
        filename : str, file-like object
            The file to write the results to.
        notes : str, list
            Any notes to be added to the report. If a string, adds everything as one line.
            If a list, must be a list of strings; each string item will be a new line.
        open_file : bool
            Whether to open the file after creation. Will use the default PDF program.
        metadata : dict
            Any data that should be appended to every page of the report. This differs from notes in that
            metadata is at the top of every page while notes is at the bottom of the report.
        """
        was_adjusted = "Yes" if self.output_was_adjusted else "No"
        title = "TRS-398 Electron Report - {} MV".format(self.energy)

        canvas = PylinacCanvas(filename, page_title=title, metadata=metadata)
        text = [
            "Site Data:",
            f"Institution: {self.institution}",
            f"Performed by: {self.physicist}",
            f"Measurement Date: {self.measurement_date}",
            f'Date of Report: {datetime.now().strftime("%A, %B %d, %Y")}',
            f"Unit: {self.unit}",
            f"Energy: {self.energy} MeV",
            f"Cone: {self.cone}",
            f"MU: {self.mu}",
            "",
            "Instrumentation:",
            f"Chamber: {self.chamber}",
            f"Chamber Calibration Factor Ndw (cGy/nC): {self.n_dw:2.3f}",
            f"Electrometer: {self.electrometer}",
            f"Kelec: {self.k_elec:2.3f}",
            "",
            "Beam Quality:",
            f"I50 (cm): {self.i_50:2.2f}",
            f"R50 (cm): {self.r_50:2.2f}",
            f"Zref (cm): {self.zref:2.2f}",
            f"Calculated kQ: {self.kq:2.3f}",
            "",
            "Chamber Corrections/Measurements:",
            f"Temperature (\N{DEGREE SIGN}C): {self.temp:2.1f}",
            f"Pressure (kPa): {self.press:2.1f}",
            f"Mraw @ ({self.voltage_reference}V, Reference) (nC): {self.m_reference}",
            f"Mraw @ ({self.voltage_reduced}V, Reduced) (nC): {self.m_reduced}",
            f"Mraw @ ({-self.voltage_reference}V, Opposite) (nC): {self.m_opposite}",
            f"Ktp: {self.k_tp:2.3f}",
            f"Ks: {self.k_s:2.3f}",
            f"Kpol: {self.k_pol:2.3f}",
            "",
            "Dose Determination:",
            f"Fully corrected M (nC): {self.m_corrected:2.3f}",
            f"Tissue correction (e.g. muscle): {self.tissue_correction:2.3f}",
            f"Dose/MU @ zref (cGy): {self.dose_mu_zref:2.3f}",
            f"Clinical PDD (%): {self.clinical_pdd_zref:2.2f}",
            f"Dose/MU @ zmax (cGy): {self.dose_mu_zmax:2.3f}",
            "",
            f"Output Adjustment?: {was_adjusted}",
        ]
        if was_adjusted == "Yes":
            text.append(
                f"Adjusted Mraw @ reference voltage (nC): {self.m_reference_adjusted}"
            )
            text.append(
                f"Adjusted fully corrected M (nC): {self.m_corrected_adjusted:2.3f}"
            )
            text.append(
                f"Adjusted Dose/MU @ zref depth (cGy): {self.dose_mu_zref_adjusted:2.3f}"
            )
            text.append(
                f"Adjusted Dose/MU @ zmax (cGy): {self.dose_mu_zmax_adjusted:2.3f}"
            )
        canvas.add_text(text=text, location=(2, 25.5), font_size=12)
        if notes is not None:
            canvas.add_text(text="Notes:", location=(12, 6.5), font_size=14)
            canvas.add_text(text=notes, location=(12, 6))

        canvas.finish()

        if open_file:
            webbrowser.open(filename)



from abc import ABC

import numpy as np
from pydicom.dataset import Dataset, FileMetaDataset

from .layers import Layer


class Simulator(ABC):
    """Abstract class for an image simulator"""

    pixel_size: float
    shape: (int, int)

    def __init__(self, sid: float = 1500):
        """

        Parameters
        ----------
        sid
            Source to image distance in mm.
        """
        self.image = np.zeros(self.shape, np.uint16)
        self.sid = sid
        self.mag_factor = sid / 1000

    def add_layer(self, layer: Layer) -> None:
        """Add a layer to the image"""
        self.image = layer.apply(self.image, self.pixel_size, self.mag_factor)

    def generate_dicom(self, file_out_name: str, **kwargs):
        """Generate a DICOM file with the constructed image (via add_layer)"""
        raise NotImplementedError(
            "This method has not been implemented for this simulator. Overload the method of your simulator."
        )


class AS500Image(Simulator):
    """Simulates an AS500 EPID image."""

    pixel_size: float = 0.78125
    shape: (int, int) = (384, 512)

    def generate_dicom(
        self,
        file_out_name: str,
        gantry_angle: float = 0.0,
        coll_angle: float = 0.0,
        table_angle: float = 0.0,
    ) -> None:
        # make image look like an EPID with flipped data (dose->low)
        flipped_image = -self.image + self.image.max() + self.image.min()

        file_meta = FileMetaDataset()
        # Main data elements
        ds = Dataset()
        ds.ImageType = ["DERIVED", "SECONDARY", "PORTAL"]
        ds.SOPClassUID = "1.2.840.10008.5.1.4.1.1.481.1"
        ds.SOPInstanceUID = "1.2.840.113854.141883099300381770008774160544352783139.1.1"
        ds.StudyDate = "20150212"
        ds.ContentDate = "20150212"
        ds.StudyTime = "124120"
        ds.ContentTime = "124120"
        ds.AccessionNumber = ""
        ds.Modality = "RTIMAGE"
        ds.ConversionType = "WSD"
        ds.Manufacturer = "IMPAC Medical Systems, Inc."
        ds.ReferringPhysicianName = ""
        ds.StudyDescription = "QA"
        ds.SeriesDescription = "D + Gantry_0"
        ds.PhysiciansOfRecord = "Awesome Physician"
        ds.OperatorsName = ""
        ds.ManufacturerModelName = "MOSAIQ"
        ds.PatientName = "Lutz^Test Tool"
        ds.PatientID = "zzzBAC_Lutz"
        ds.PatientBirthDate = ""
        ds.SoftwareVersions = "2.41.01J0"
        ds.StudyInstanceUID = "1.2.840.113854.141883099300381770008774160544352783139"
        ds.SeriesInstanceUID = (
            "1.2.840.113854.141883099300381770008774160544352783139.1"
        )
        ds.StudyID = "348469"
        ds.SeriesNumber = "4597199"
        ds.InstanceNumber = "0"
        ds.PatientOrientation = ""
        ds.PositionReferenceIndicator = ""
        ds.SamplesPerPixel = 1
        ds.PhotometricInterpretation = "MONOCHROME2"
        ds.Rows = self.image.shape[0]
        ds.Columns = self.image.shape[1]
        ds.BitsAllocated = 16
        ds.BitsStored = 16
        ds.HighBit = 15
        ds.PixelRepresentation = 0
        ds.RTImageLabel = "D"
        ds.RTImagePlane = "NORMAL"
        ds.XRayImageReceptorAngle = "0.0"
        ds.ImagePlanePixelSpacing = [self.pixel_size, self.pixel_size]
        ds.RTImagePosition = [-200.70400, 150.52800]
        ds.RadiationMachineSAD = "1000.0"
        ds.RTImageSID = self.sid
        ds.PrimaryDosimeterUnit = "MU"
        ds.GantryAngle = str(gantry_angle)
        ds.BeamLimitingDeviceAngle = str(coll_angle)
        ds.PatientSupportAngle = str(table_angle)
        ds.PixelData = flipped_image  # XXX Array of 393216 bytes excluded

        ds.file_meta = file_meta
        ds.is_implicit_VR = True
        ds.is_little_endian = True
        ds.save_as(file_out_name, write_like_original=False)


class AS1000Image(Simulator):
    """Simulates an AS1000 EPID image."""

    pixel_size: float = 0.390625
    shape: (int, int) = (768, 1024)

    def generate_dicom(
        self,
        file_out_name: str,
        gantry_angle: float = 0.0,
        coll_angle: float = 0.0,
        table_angle: float = 0.0,
    ) -> None:
        # make image look like an EPID with flipped data (dose->low)
        flipped_image = -self.image + self.image.max() + self.image.min()

        # File meta info data elements
        file_meta = FileMetaDataset()

        # Main data elements
        ds = Dataset()
        ds.ImageType = ["DERIVED", "SECONDARY", "PORTAL"]
        ds.SOPClassUID = "1.2.840.10008.5.1.4.1.1.481.1"
        ds.SOPInstanceUID = "1.2.840.113854.323870129946883845737820671794195198978.1.1"
        ds.StudyDate = "20140819"
        ds.ContentDate = "20140819"
        ds.StudyTime = "130944"
        ds.ContentTime = "130944"
        ds.AccessionNumber = ""
        ds.Modality = "RTIMAGE"
        ds.ConversionType = "WSD"
        ds.Manufacturer = "IMPAC Medical Systems, Inc."
        ds.ReferringPhysicianName = ""
        ds.StudyDescription = "QA"
        ds.SeriesDescription = "Q + Couch_270"
        ds.PhysiciansOfRecord = "I am king"
        ds.OperatorsName = ""
        ds.ManufacturerModelName = "MOSAIQ"
        ds.PatientName = "Albert Einstein"
        ds.PatientID = "abc123"
        ds.PatientBirthDate = ""
        ds.SoftwareVersions = "2.41.01J0"
        ds.StudyInstanceUID = "1.2.840.113854.323870129946883845737820671794195198978"
        ds.SeriesInstanceUID = (
            "1.2.840.113854.323870129946883845737820671794195198978.1"
        )
        ds.StudyID = "348469"
        ds.SeriesNumber = "4290463"
        ds.InstanceNumber = "0"
        ds.PatientOrientation = ""
        ds.PositionReferenceIndicator = ""
        ds.SamplesPerPixel = 1
        ds.PhotometricInterpretation = "MONOCHROME2"
        ds.Rows = self.shape[0]
        ds.Columns = self.shape[1]
        ds.BitsAllocated = 16
        ds.BitsStored = 16
        ds.HighBit = 15
        ds.PixelRepresentation = 0
        ds.RTImageLabel = "Q"
        ds.RTImagePlane = "NORMAL"
        ds.XRayImageReceptorAngle = "0.0"
        ds.ImagePlanePixelSpacing = [self.pixel_size, self.pixel_size]
        ds.RTImagePosition = [-200.70400, 150.523400]
        ds.RadiationMachineSAD = "1000.0"
        ds.RTImageSID = self.sid
        ds.PrimaryDosimeterUnit = "MU"
        ds.GantryAngle = str(gantry_angle)
        ds.BeamLimitingDeviceAngle = str(coll_angle)
        ds.PatientSupportAngle = str(table_angle)
        # ds.CurveDimensions = 2
        # ds.NumberOfPoints = 4
        # ds.TypeOfData = 'ROI'
        # ds.CurveDescription = 'VContour 30'
        # ds.AxisUnits = ['PIXL', 'PIXL']
        # ds.DataValueRepresentation = 3
        # ds.CurveLabel = 'Field Edge (Q:MV)'
        # ds.CurveData = b'\x00\x00\x00\x00 .\x81@\x00\x00\x00\x00\x10\\z@\x00\x00\x00\x00 .\x81@\x00\x00\x00\x00\x90\x93u@\x00\x00\x00\x00\xc0\x93}@\x00\x00\x00\x00\x90\x93u@\x00\x00\x00\x00\xc0\x93}@\x00\x00\x00\x00\x10\\z@'
        ds.PixelData = flipped_image  # XXX Array of 1572864 bytes excluded

        ds.file_meta = file_meta
        ds.is_implicit_VR = True
        ds.is_little_endian = True
        ds.save_as(file_out_name, write_like_original=False)


class AS1200Image(Simulator):
    """Simulates an AS1200 EPID image."""

    pixel_size: float = 0.336
    shape: (int, int) = (1280, 1280)

    def generate_dicom(
        self,
        file_out_name: str,
        gantry_angle: float = 0.0,
        coll_angle: float = 0.0,
        table_angle: float = 0.0,
    ) -> None:
        file_meta = FileMetaDataset()
        file_meta.FileMetaInformationGroupLength = 196
        file_meta.FileMetaInformationVersion = b"\x00\x01"
        file_meta.MediaStorageSOPClassUID = "1.2.840.10008.5.1.4.1.1.481.1"
        file_meta.MediaStorageSOPInstanceUID = (
            "1.2.246.352.64.1.5468686515961995030.4457606667843517571"
        )
        file_meta.TransferSyntaxUID = "1.2.840.10008.1.2"
        file_meta.ImplementationClassUID = "1.2.246.352.70.2.1.120.1"
        file_meta.ImplementationVersionName = "MergeCOM3_410"

        # Main data elements
        ds = Dataset()
        ds.SpecificCharacterSet = "ISO_IR 100"
        ds.ImageType = ["ORIGINAL", "PRIMARY", "PORTAL"]
        ds.InstanceCreationDate = "20161230"
        ds.InstanceCreationTime = "215510"
        ds.SOPClassUID = "1.2.840.10008.5.1.4.1.1.481.1"
        ds.SOPInstanceUID = "1.2.246.352.64.1.5468686515961995030.4457606667843517571"
        ds.StudyDate = "20161230"
        ds.ContentDate = "20161230"
        ds.StudyTime = "215441.936"
        ds.ContentTime = "215441.936"
        ds.AccessionNumber = ""
        ds.Modality = "RTIMAGE"
        ds.ConversionType = ""
        ds.Manufacturer = "Varian Medical Systems"
        ds.ReferringPhysicianName = ""
        ds.StationName = "NDS-WKS-SN9999"
        ds.OperatorsName = "King Kong"
        ds.ManufacturerModelName = "VMS.XI.Service"
        ds.PatientName = "Grace Hopper"
        ds.PatientID = "VMS.XI.Service"
        ds.PatientBirthDate = "19000101"
        ds.PatientBirthTime = "000000"
        ds.PatientSex = ""
        ds.SoftwareVersions = "2.5.13.2"
        ds.StudyInstanceUID = "1.2.246.352.64.4.5644626269434644263.1905029945372990626"
        ds.SeriesInstanceUID = (
            "1.2.246.352.64.2.5508761605912087323.11665958260371685307"
        )
        ds.StudyID = "fdd794f2-8520-4c4a-aecc-e4446c1730ff"
        ds.SeriesNumber = None
        ds.AcquisitionNumber = "739774555"
        ds.InstanceNumber = "1"
        ds.PatientOrientation = ""
        ds.FrameOfReferenceUID = (
            "1.2.246.352.64.3.4714322356925391886.9391210174715030407"
        )
        ds.PositionReferenceIndicator = ""
        ds.SamplesPerPixel = 1
        ds.PhotometricInterpretation = "MONOCHROME1"
        ds.PlanarConfiguration = 0
        ds.Rows = self.shape[0]
        ds.Columns = self.shape[1]
        ds.BitsAllocated = 16
        ds.BitsStored = 16
        ds.HighBit = 15
        ds.PixelRepresentation = 0
        ds.WindowCenter = "32767.0"
        ds.WindowWidth = "65535.0"
        ds.RescaleIntercept = "0.0"
        ds.RescaleSlope = "1.0"
        ds.RescaleType = "US"
        ds.RTImageLabel = "MV_180"
        ds.RTImageDescription = ""
        ds.ReportedValuesOrigin = "ACTUAL"
        ds.RTImagePlane = "NORMAL"
        ds.XRayImageReceptorTranslation = [0.00, 0.00, 1000 - self.sid]
        ds.XRayImageReceptorAngle = "0.0"
        ds.RTImageOrientation = [1, 0, 0, 0, -1, 0]
        ds.ImagePlanePixelSpacing = [self.pixel_size, self.pixel_size]
        ds.RTImagePosition = [-214.872, 214.872]
        ds.RadiationMachineName = "TrueBeam from Hell"
        ds.RadiationMachineSAD = "1000.0"
        ds.RTImageSID = self.sid
        ds.PrimaryDosimeterUnit = "MU"
        ds.GantryAngle = str(gantry_angle)
        ds.BeamLimitingDeviceAngle = str(coll_angle)
        ds.PatientSupportAngle = str(table_angle)
        ds.TableTopVerticalPosition = "-24.59382842824"
        ds.TableTopLongitudinalPosition = "200.813502948597"
        ds.TableTopLateralPosition = "3.00246706215532"
        ds.PixelData = self.image  # XXX Array of 3276800 bytes excluded

        ds.file_meta = file_meta
        ds.is_implicit_VR = True
        ds.is_little_endian = True
        ds.save_as(file_out_name, write_like_original=False)



"""Utility functions for pylinac."""
import os
import os.path as osp
import struct
from collections.abc import Iterable
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Union, Sequence, Type, BinaryIO

import numpy as np
import pydicom

from .. import __version__


def convert_to_enum(value: Union[str, Enum, None], enum: Type[Enum]) -> Enum:
    """Convert a value to an enum representation from an enum value if needed"""
    if isinstance(value, enum):
        return value
    else:
        return enum(value)


@dataclass
class ResultBase:
    pylinac_version: str = field(init=False)  #:
    date_of_analysis: datetime = field(init=False)  #:

    def __post_init__(self):
        self.pylinac_version = __version__
        self.date_of_analysis = datetime.today()


def clear_data_files():
    """Delete all demo files, image classifiers, etc from the demo folder"""
    demo_folder = osp.join(osp.dirname(osp.dirname(__file__)), "demo_files")
    if osp.isdir(demo_folder):
        for file in os.listdir(demo_folder):
            full_file = osp.join(demo_folder, file)
            if osp.isfile(full_file):
                os.remove(full_file)
    print("Pylinac data files cleared.")


def assign2machine(source_file: str, machine_file: str):
    """Assign a DICOM RT Plan file to a specific machine. The source file is overwritten to contain
    the machine of the machine file.

    Parameters
    ----------
    source_file : str
        Path to the DICOM RTPlan file that contains the fields/plan desired
        (e.g. a Winston Lutz set of fields or Varian's default PF files).
    machine_file : str
        Path to a DICOM RTPlan file that has the desired machine. This is easily obtained from pushing a plan from the TPS
        for that specific machine. The file must contain at least one valid field.
    """
    dcm_source = pydicom.dcmread(source_file)
    dcm_machine = pydicom.dcmread(machine_file)
    for beam in dcm_source.BeamSequence:
        beam.TreatmentMachineName = dcm_machine.BeamSequence[0].TreatmentMachineName
    dcm_source.save_as(source_file)


def is_close(
    val: float, target: Union[float, Sequence], delta: float = 1
):
    """Return whether the value is near the target value(s).

    Parameters
    ----------
    val : number
        The value being compared against.
    target : number, iterable
        If a number, the values are simply evaluated.
        If a sequence, each target is compared to ``val``.
        If any values of ``target`` are close, the comparison is considered True.

    Returns
    -------
    bool
    """
    try:
        targets = (value for value in target)
    except (AttributeError, TypeError):
        targets = [target]
    for target in targets:
        if target - delta < val < target + delta:
            return True
    return False


def simple_round(number: float, decimals: int = 0) -> float:
    """Round a number to the given number of decimals. Fixes small floating number errors."""
    num = int(round(number * 10**decimals))
    num /= 10**decimals
    return num


def is_iterable(object) -> bool:
    """Determine if an object is iterable."""
    return isinstance(object, Iterable)


class Structure:
    """A simple structure that assigns the arguments to the object."""

    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)

    def update(self, **kwargs):
        self.__dict__.update(**kwargs)


def decode_binary(
    file: BinaryIO,
    dtype: Union[Type[int], Type[float], Type[str]],
    num_values: int = 1,
    cursor_shift: int = 0,
) -> Union[int, float, str, np.ndarray]:
    """Read in a raw binary file and convert it to given data types.

    Parameters
    ----------
    file
        The open file object.
    dtype
        The expected data type to return. If int or float and num_values > 1, will return numpy array.
    num_values
        The expected number of dtype to return

        .. note:: This is not the same as the number of bytes.

    cursor_shift : int
        The number of bytes to move the cursor forward after decoding. This is used if there is a
        reserved section after the read-in segment.
    """
    f = file

    if dtype == str:  # if string
        output = f.read(num_values)
        if type(f) is not str:  # in py3 fc will be bytes
            output = output.decode()
        # strip the padding ("\x00")
        output = output.strip("\x00")
    elif dtype == int:
        ssize = struct.calcsize("i") * num_values
        output = np.asarray(struct.unpack("i" * num_values, f.read(ssize)))
        if len(output) == 1:
            output = int(output)
    elif dtype == float:
        ssize = struct.calcsize("f") * num_values
        output = np.asarray(struct.unpack("f" * num_values, f.read(ssize)))
        if len(output) == 1:
            output = float(output)
    else:
        raise TypeError(f"datatype '{dtype}' was not valid")

    # shift cursor if need be (e.g. if a reserved section follows)
    if cursor_shift:
        f.seek(cursor_shift, 1)
    return output
